"""Utility functions for the extract module."""
from __future__ import division

import logging
import os
import os.path as op

import numpy as np
import pandas as pd
import requests
from fuzzywuzzy import fuzz

from ..utils import uk_to_us

LGR = logging.getLogger(__name__)


def get_data_dirs(data_dir=None):
    """Return the directories in which NiMARE looks for data.

    This is typically useful for the end-user to check where the data is
    downloaded and stored.

    Parameters
    ----------
    data_dir: string, optional
        Path of the data directory. Used to force data storage in a specified
        location. Default: None

    Returns
    -------
    paths: list of strings
        Paths of the dataset directories.

    Notes
    -----
    Taken from Nilearn.
    This function retrieves the datasets directories using the following
    priority :
    1. defaults system paths
    2. the keyword argument data_dir
    3. the global environment variable NIMARE_SHARED_DATA
    4. the user environment variable NIMARE_DATA
    5. nimare_data in the user home folder
    """
    # We build an array of successive paths by priority
    # The boolean indicates if it is a pre_dir: in that case, we won't add the
    # dataset name to the path.
    paths = []

    # Check data_dir which force storage in a specific location
    if data_dir is not None:
        paths.extend(data_dir.split(os.pathsep))

    # If data_dir has not been specified, then we crawl default locations
    if data_dir is None:
        global_data = os.getenv("NIMARE_SHARED_DATA")
        if global_data is not None:
            paths.extend(global_data.split(os.pathsep))

        local_data = os.getenv("NIMARE_DATA")
        if local_data is not None:
            paths.extend(local_data.split(os.pathsep))

        paths.append(os.path.expanduser("~/.nimare"))
    return paths


def _get_dataset_dir(dataset_name, data_dir=None, default_paths=None, verbose=1):
    """Create if necessary and returns data directory of given dataset.

    Parameters
    ----------
    dataset_name: string
        The unique name of the dataset.
    data_dir: string, optional
        Path of the data directory. Used to force data storage in a specified
        location. Default: None
    default_paths: list of string, optional
        Default system paths in which the dataset may already have been
        installed by a third party software. They will be checked first.
    verbose: int, optional
        verbosity level (0 means no message).

    Returns
    -------
    data_dir: string
        Path of the given dataset directory.

    Notes
    -----
    Taken from Nilearn.
    This function retrieves the datasets directory (or data directory) using
    the following priority :
    1. defaults system paths
    2. the keyword argument data_dir
    3. the global environment variable NIMARE_SHARED_DATA
    4. the user environment variable NIMARE_DATA
    5. nimare_data in the user home folder
    """
    paths = []
    # Search possible data-specific system paths
    if default_paths is not None:
        for default_path in default_paths:
            paths.extend([(d, True) for d in default_path.split(os.pathsep)])

    paths.extend([(d, False) for d in get_data_dirs(data_dir=data_dir)])

    if verbose > 2:
        print("Dataset search paths: %s" % paths)

    # Check if the dataset exists somewhere
    for path, is_pre_dir in paths:
        if not is_pre_dir:
            path = os.path.join(path, dataset_name)
        if os.path.islink(path):
            # Resolve path
            path = readlinkabs(path)
        if os.path.exists(path) and os.path.isdir(path):
            if verbose > 1:
                print("\nDataset found in %s\n" % path)
            return path

    # If not, create a folder in the first writeable directory
    errors = []
    for (path, is_pre_dir) in paths:
        if not is_pre_dir:
            path = os.path.join(path, dataset_name)
        if not os.path.exists(path):
            try:
                os.makedirs(path)
                if verbose > 0:
                    print("\nDataset created in %s\n" % path)
                return path
            except Exception as exc:
                short_error_message = getattr(exc, "strerror", str(exc))
                errors.append("\n -{0} ({1})".format(path, short_error_message))

    raise OSError(
        "NiMARE tried to store the dataset in the following directories, but: " + "".join(errors)
    )


def readlinkabs(link):
    """Return an absolute path for the destination of a symlink.

    From nilearn.
    """
    path = os.readlink(link)
    if os.path.isabs(path):
        return path
    return os.path.join(os.path.dirname(link), path)


def _download_zipped_file(url, filename=None):
    """Download from a URL to a file."""
    if filename is None:
        data_dir = op.abspath(op.getcwd())
        filename = op.join(data_dir, url.split("/")[-1])
    # NOTE the stream=True parameter
    req = requests.get(url, stream=True)
    with open(filename, "wb") as f_obj:
        for chunk in req.iter_content(chunk_size=1024):
            if chunk:  # filter out keep-alive new chunks
                f_obj.write(chunk)
    return filename


def _longify(df):
    """Expand comma-separated lists of aliases in DataFrame into separate rows."""
    reduced = df[["id", "name", "alias"]]
    rows = []
    for index, row in reduced.iterrows():
        if isinstance(row["alias"], str) and "," in row["alias"]:
            aliases = row["alias"].split(", ") + [row["name"]]
        else:
            aliases = [row["name"]]

        for alias in aliases:
            rows.append([row["id"], row["name"].lower(), alias.lower()])
    out_df = pd.DataFrame(columns=["id", "name", "alias"], data=rows)
    out_df = out_df.replace("", np.nan)
    return out_df


def _get_ratio(tup):
    """Get fuzzy ratio."""
    if all(isinstance(t, str) for t in tup):
        return fuzz.ratio(tup[0], tup[1])
    else:
        return 100


def _gen_alt_forms(term):
    """Generate a list of alternate forms for a given term."""
    if not isinstance(term, str) or len(term) == 0:
        return [None]

    alt_forms = []
    # For one alternate form, put contents of parentheses at beginning of term
    if "(" in term:
        prefix = term[term.find("(") + 1 : term.find(")")]
        temp_term = term.replace("({0})".format(prefix), "").replace("  ", " ")
        alt_forms.append(temp_term)
        alt_forms.append("{0} {1}".format(prefix, temp_term))
    else:
        prefix = ""

    # Remove extra spaces
    alt_forms = [s.strip() for s in alt_forms]

    # Allow plurals
    # temp = [s+'s' for s in alt_forms]
    # temp += [s+'es' for s in alt_forms]
    # alt_forms += temp

    # Remove words "task" and/or "paradigm"
    alt_forms += [term.replace(" task", "") for term in alt_forms]
    alt_forms += [term.replace(" paradigm", "") for term in alt_forms]

    # Remove duplicates
    alt_forms = list(set(alt_forms))
    return alt_forms


def _get_concept_reltype(relationship, direction):
    """Convert two-part relationship info to more parsimonious representation.

    The two part representation includes relationship type and direction.
    """
    new_rel = None
    if relationship == "PARTOF":
        if direction == "child":
            new_rel = "hasPart"
        elif direction == "parent":
            new_rel = "isPartOf"
    elif relationship == "KINDOF":
        if direction == "child":
            new_rel = "hasKind"
        elif direction == "parent":
            new_rel = "isKindOf"
    return new_rel


def _expand_df(df):
    """Add alternate forms to DataFrame, then sort DataFrame by alias length and similarity.

    Sorting by alias length is done for order of extraction from text. Sorting by similarity to
    original name is done in order to select most appropriate term to associate with alias.
    """
    df = df.copy()
    df["alias"] = df["alias"].apply(uk_to_us)
    new_rows = []
    for index, row in df.iterrows():
        alias = row["alias"]
        alt_forms = _gen_alt_forms(alias)
        for alt_form in alt_forms:
            temp_row = row.copy()
            temp_row["alias"] = alt_form
            new_rows.append(temp_row.tolist())
    alt_df = pd.DataFrame(columns=df.columns, data=new_rows)
    df = pd.concat((df, alt_df), axis=0)
    # Sort by name length and similarity of alternate form to preferred term
    # For example, "task switching" the concept should take priority over the
    # "task switching" version of the "task-switching" task.
    df["length"] = df["alias"].str.len()
    df["ratio"] = df[["alias", "name"]].apply(_get_ratio, axis=1)
    df = df.sort_values(by=["length", "ratio"], ascending=[False, False])
    return df



"""Utility functions for decoding/encoding."""
import numpy as np


def weight_priors(topic_priors, prior_weight):
    """
    Combine topic priors with prior weight.

    Parameters
    ----------
    topic_priors : array-like
        The prior weights for topics (n_topics-long array). Scale may be
        arbitrary, as the array will be normalized.
    prior_weight : :obj:`float`
        Scalar by which to weight priors.

    Returns
    -------
    weighted_priors : :obj:`numpy.ndarray`
        Updated prior weights for topics.
    """
    if not isinstance(prior_weight, (float, int)):
        raise IOError("Input prior_weight must be a float in range (0, 1)")
    elif not 0.0 <= prior_weight <= 1:
        raise ValueError("Input prior_weight must be in range (0, 1)")

    # Enforce compatible types
    topic_priors = topic_priors.astype(float)
    prior_weight = float(prior_weight)

    # Normalize priors
    topic_priors /= np.sum(topic_priors)

    # Weight priors
    topic_priors *= prior_weight

    # Create uniform distribution to combine with priors
    uniform = np.ones(topic_priors.shape)
    uniform /= np.sum(uniform)
    uniform *= 1 - prior_weight

    # Weight priors with uniform base
    weighted_priors = topic_priors + uniform
    return weighted_priors



import importlib.util
import json
import os.path as op

spec = importlib.util.spec_from_file_location(
    "_version", op.join(op.dirname(__file__), "nimare/_version.py")
)
_version = importlib.util.module_from_spec(spec)
spec.loader.exec_module(_version)

VERSION = _version.get_versions()["version"]
del _version

# Get list of authors from Zenodo file
with open(op.join(op.dirname(__file__), ".zenodo.json"), "r") as fo:
    zenodo_info = json.load(fo)
authors = [author["name"] for author in zenodo_info["creators"]]
authors = [author.split(", ")[1] + " " + author.split(", ")[0] for author in authors]

AUTHOR = "NiMARE developers"
COPYRIGHT = "Copyright 2018--, NiMARE developers"
CREDITS = authors
LICENSE = "MIT"
MAINTAINER = "Taylor Salo"
EMAIL = "tsalo006@fiu.edu"
STATUS = "Prototype"
URL = "https://github.com/neurostuff/NiMARE"
PACKAGENAME = "NiMARE"
DESCRIPTION = "NiMARE: Neuroimaging Meta-Analysis Research Environment"
LONGDESC = """
NiMARE
======
NiMARE (Neuroimaging Meta-Analysis Research Environment) is a Python package
for coordinate-based and image-based meta-analysis of neuroimaging data.

License
=======
``NiMARE`` is licensed under the terms of the MIT license. See the file
'LICENSE' for information on the history of this software, terms & conditions
for usage, and a DISCLAIMER OF ALL WARRANTIES.

All trademarks referenced herein are property of their respective holders.

Copyright (c) 2018--, NiMARE developers
"""

DOWNLOAD_URL = "https://github.com/neurostuff/{name}/archive/{ver}.tar.gz".format(
    name=PACKAGENAME, ver=VERSION
)

REQUIRES = [
    "cognitiveatlas",
    "fuzzywuzzy",
    "indexed_gzip>=1.4.0",
    "matplotlib",
    "nibabel>=3.0.0",
    "nilearn>=0.7.1",
    "numpy",
    "pandas",
    "pymare>=0.0.2",
    "scikit-learn",
    "scipy",
    "seaborn",
    "statsmodels",
    "tqdm",
]

TESTS_REQUIRES = [
    "codecov",
    "coverage",
    "coveralls",
    "flake8-black",
    "flake8-docstrings",
    "flake8-isort",
    "pytest",
    "pytest-cov",
]

EXTRA_REQUIRES = {
    "peaks2maps-cpu": ["tensorflow>=2.0.0", "appdirs"],
    "peaks2maps-gpu": ["tensorflow-gpu>=2.0.0", "appdirs"],
    "doc": [
        "m2r",
        "pillow",
        "recommonmark",
        "sphinx>=3.1.1",
        "sphinx-argparse",
        "sphinx-copybutton",
        "sphinx_gallery",
        "sphinx_rtd_theme",
    ],
    "tests": TESTS_REQUIRES,
    "duecredit": ["duecredit"],
}

# Enable a handle to install all extra dependencies at once
EXTRA_REQUIRES["all"] = list(set([v for deps in EXTRA_REQUIRES.values() for v in deps]))

ENTRY_POINTS = {"console_scripts": ["nimare=nimare.cli:_main"]}

# Package classifiers
CLASSIFIERS = [
    "Development Status :: 3 - Alpha",
    "Environment :: Console",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Topic :: Scientific/Engineering",
]



"""Kernel transformers for CBMA algorithms.

Methods for estimating thresholded cluster maps from neuroimaging contrasts
(Contrasts) from sets of foci and optional additional information (e.g., sample
size and test statistic values).
"""
from __future__ import division

import logging
import os
from hashlib import md5

import nibabel as nib
import numpy as np
import pandas as pd
from nilearn import image

from ..base import Transformer
from ..utils import use_memmap, vox2mm
from .utils import compute_ale_ma, compute_kda_ma, compute_p2m_ma, get_ale_kernel

LGR = logging.getLogger(__name__)


class KernelTransformer(Transformer):
    """Base class for modeled activation-generating methods in :mod:`nimare.meta.kernel`.

    Coordinate-based meta-analyses leverage coordinates reported in
    neuroimaging papers to simulate the thresholded statistical maps from the
    original analyses. This generally involves convolving each coordinate with
    a kernel (typically a Gaussian or binary sphere) that may be weighted based
    on some additional measure, such as statistic value or sample size.

    Notes
    -----
    All extra (non-ijk) parameters for a given kernel should be overrideable as
    parameters to __init__, so we can access them with get_params() and also
    apply them to datasets with missing data.
    """

    def _infer_names(self, **kwargs):
        """Determine filename pattern and image type for files created with this transformer.

        The parameters used to construct the filenames come from the transformer's
        parameters (attributes saved in ``__init__()``).

        Parameters
        ----------
        **kwargs
            Additional key/value pairs to incorporate into the image name.
            A common example is the hash for the target template's affine.

        Attributes
        ----------
        filename_pattern : str
            Filename pattern for images that will be saved by the transformer.
        image_type : str
            Name of the corresponding column in the Dataset.images DataFrame.
        """
        params = self.get_params()
        params = dict(**params, **kwargs)

        # Determine names for kernel-specific files
        keys = sorted(params.keys())
        param_str = "_".join("{k}-{v}".format(k=k, v=str(params[k])) for k in keys)
        self.filename_pattern = (
            "study-[[id]]_{ps}_{n}.nii.gz".format(n=self.__class__.__name__, ps=param_str)
            .replace("[[", "{")
            .replace("]]", "}")
        )
        self.image_type = "{ps}_{n}".format(n=self.__class__.__name__, ps=param_str)

    @use_memmap(LGR)
    def transform(self, dataset, masker=None, return_type="image"):
        """Generate modeled activation images for each Contrast in dataset.

        Parameters
        ----------
        dataset : :obj:`nimare.dataset.Dataset` or :obj:`pandas.DataFrame`
            Dataset for which to make images. Can be a DataFrame if necessary.
        masker : img_like, optional
            Only used if dataset is a DataFrame.
        return_type : {'array', 'image', 'dataset'}, optional
            Whether to return a numpy array ('array'), a list of niimgs
            ('image'), or a Dataset with MA images saved as files ('dataset').
            Default is 'dataset'.

        Returns
        -------
        imgs : (C x V) :class:`numpy.ndarray` or :obj:`list` of
               :class:`nibabel.Nifti1Image` or :class:`nimare.dataset.Dataset`
            If return_type is 'array', a 2D numpy array (C x V), where C is
            contrast and V is voxel.
            If return_type is 'image', a list of modeled activation images
            (one for each of the Contrasts in the input dataset).
            If return_type is 'dataset', a new Dataset object with modeled
            activation images saved to files and referenced in the
            Dataset.images attribute.

        Attributes
        ----------
        filename_pattern : str
            Filename pattern for MA maps that will be saved by the transformer.
        image_type : str
            Name of the corresponding column in the Dataset.images DataFrame.
        """
        if return_type not in ("array", "image", "dataset"):
            raise ValueError('Argument "return_type" must be "image", "array", or "dataset".')

        if isinstance(dataset, pd.DataFrame):
            assert (
                masker is not None
            ), 'Argument "masker" must be provided if dataset is a DataFrame'
            mask = masker.mask_img
            coordinates = dataset.copy()
            assert (
                return_type != "dataset"
            ), "Input dataset must be a Dataset if return_type='dataset'."
        else:
            masker = dataset.masker if not masker else masker
            mask = masker.mask_img
            coordinates = dataset.coordinates

            # Determine MA map filenames. Must happen after parameters are set.
            self._infer_names(affine=md5(mask.affine).hexdigest())

            # Check for existing MA maps
            # Use coordinates to get IDs instead of Dataset.ids bc of possible
            # mismatch between full Dataset and contrasts with coordinates.
            if self.image_type in dataset.images.columns:
                files = dataset.get_images(ids=coordinates["id"].unique(), imtype=self.image_type)
                if all(f is not None for f in files):
                    LGR.debug("Files already exist. Using them.")
                    if return_type == "array":
                        return masker.transform(files)
                    elif return_type == "image":
                        return [nib.load(f) for f in files]
                    elif return_type == "dataset":
                        return dataset.copy()

        # Otherwise, generate the MA maps
        if return_type == "array":
            mask_data = mask.get_fdata().astype(np.bool)
        elif return_type == "image":
            dtype = type(self.value) if hasattr(self, "value") else float
            mask_data = mask.get_fdata().astype(dtype)
        elif return_type == "dataset":
            dataset = dataset.copy()
            if dataset.basepath is None:
                raise ValueError(
                    "Dataset output path is not set. Set the path with Dataset.update_path()."
                )
            elif not os.path.isdir(dataset.basepath):
                raise ValueError(
                    "Output directory does not exist. Set the path to an "
                    "existing folder with Dataset.update_path()."
                )

        transformed_maps = self._transform(mask, coordinates)

        if not isinstance(transformed_maps[0], (list, tuple)):
            if return_type == "array":
                return transformed_maps[0][:, mask_data]
            else:
                transformed_maps = list(zip(*transformed_maps))

        imgs = []
        for (kernel_data, id_) in transformed_maps:
            if return_type == "array":
                img = kernel_data[mask_data]
                imgs.append(img)
            elif return_type == "image":
                kernel_data *= mask_data
                img = nib.Nifti1Image(kernel_data, mask.affine)
                imgs.append(img)
            elif return_type == "dataset":
                img = nib.Nifti1Image(kernel_data, mask.affine)
                out_file = os.path.join(dataset.basepath, self.filename_pattern.format(id=id_))
                img.to_filename(out_file)
                dataset.images.loc[dataset.images["id"] == id_, self.image_type] = out_file

        del kernel_data

        if return_type == "array":
            return np.vstack(imgs)
        elif return_type == "image":
            del transformed_maps
            return imgs
        elif return_type == "dataset":
            del transformed_maps
            # Infer relative path
            dataset.images = dataset.images
            return dataset

    def _transform(self, mask, coordinates):
        """Apply the kernel's unique transformer.

        Parameters
        ----------
        mask : niimg-like
            Mask image. Should contain binary-like integer data.
        coordinates : pandas.DataFrame
            DataFrame containing IDs and coordinates.
            The DataFrame must have the following columns: "id", "i", "j", "k".
            Additionally, individual kernels may require other columns
            (e.g., "sample_size" for ALE).

        Returns
        -------
        transformed_maps : list of (3D array, str) tuples
            Transformed data, containing one element for each study.
            Each element is composed of a 3D array (the MA map) and the study's ID.
        """
        pass


class ALEKernel(KernelTransformer):
    """Generate ALE modeled activation images from coordinates and sample size.

    Parameters
    ----------
    fwhm : :obj:`float`, optional
        Full-width half-max for Gaussian kernel, if you want to have a
        constant kernel across Contrasts. Mutually exclusive with
        ``sample_size``.
    sample_size : :obj:`int`, optional
        Sample size, used to derive FWHM for Gaussian kernel based on
        formulae from Eickhoff et al. (2012). This sample size overwrites
        the Contrast-specific sample sizes in the dataset, in order to hold
        kernel constant across Contrasts. Mutually exclusive with ``fwhm``.
    low_memory : :obj:`bool`, optional
        Whether to employ mem-mapped arrays to reduce memory usage or not.
        Default=False.
    """

    def __init__(self, fwhm=None, sample_size=None, low_memory=False):
        if fwhm is not None and sample_size is not None:
            raise ValueError('Only one of "fwhm" and "sample_size" may be provided.')
        self.fwhm = fwhm
        self.sample_size = sample_size
        self.low_memory = low_memory

    def _transform(self, mask, coordinates):
        kernels = {}  # retain kernels in dictionary to speed things up
        exp_ids = coordinates["id"].unique()

        if self.low_memory:
            # Use a memmapped 4D array
            transformed_shape = (len(exp_ids),) + mask.shape
            transformed = np.memmap(
                self.memmap_filenames[0],
                dtype=float,
                mode="w+",
                shape=transformed_shape,
            )
        else:
            # Use a list of tuples
            transformed = []

        for i_exp, id_ in enumerate(exp_ids):
            data = coordinates.loc[coordinates["id"] == id_]

            ijk = np.vstack((data.i.values, data.j.values, data.k.values)).T.astype(int)
            if self.sample_size is not None:
                sample_size = self.sample_size
            elif self.fwhm is None:
                sample_size = data.sample_size.astype(float).values[0]

            if self.fwhm is not None:
                assert np.isfinite(self.fwhm), "FWHM must be finite number"
                if self.fwhm not in kernels.keys():
                    _, kern = get_ale_kernel(mask, fwhm=self.fwhm)
                    kernels[self.fwhm] = kern
                else:
                    kern = kernels[self.fwhm]
            else:
                assert np.isfinite(sample_size), "Sample size must be finite number"
                if sample_size not in kernels.keys():
                    _, kern = get_ale_kernel(mask, sample_size=sample_size)
                    kernels[sample_size] = kern
                else:
                    kern = kernels[sample_size]
            kernel_data = compute_ale_ma(mask.shape, ijk, kern)

            if self.low_memory:
                transformed[i_exp, :, :, :] = kernel_data

                # Write changes to disk
                transformed.flush()
            else:
                transformed.append((kernel_data, id_))

        if self.low_memory:
            return transformed, exp_ids
        else:
            return transformed


class KDAKernel(KernelTransformer):
    """Generate KDA modeled activation images from coordinates.

    Parameters
    ----------
    r : :obj:`int`, optional
        Sphere radius, in mm.
    value : :obj:`int`, optional
        Value for sphere.
    low_memory : :obj:`bool`, optional
        Whether to employ mem-mapped arrays to reduce memory usage or not.
        Default=False.
    """

    _sum_overlap = True

    def __init__(self, r=10, value=1, low_memory=False):
        self.r = float(r)
        self.value = value
        self.low_memory = low_memory

    def _transform(self, mask, coordinates):
        dims = mask.shape
        vox_dims = mask.header.get_zooms()

        ijks = coordinates[["i", "j", "k"]].values
        exp_idx = coordinates["id"].values
        transformed = compute_kda_ma(
            dims,
            vox_dims,
            ijks,
            self.r,
            self.value,
            exp_idx,
            sum_overlap=self._sum_overlap,
            memmap_filename=self.memmap_filenames[0],
        )
        exp_ids = np.unique(exp_idx)
        return transformed, exp_ids


class MKDAKernel(KDAKernel):
    """Generate MKDA modeled activation images from coordinates.

    Parameters
    ----------
    r : :obj:`int`, optional
        Sphere radius, in mm.
    value : :obj:`int`, optional
        Value for sphere.
    low_memory : :obj:`bool`, optional
        Whether to employ mem-mapped arrays to reduce memory usage or not.
        Default=False.
    """

    _sum_overlap = False


class Peaks2MapsKernel(KernelTransformer):
    """Generate peaks2maps modeled activation images from coordinates.

    Parameters
    ----------
    resample_to_mask : :obj:`bool`, optional
        If True, will resample the MA maps to the mask's header.
        Default is True.

    Warning
    -------
    Peaks2MapsKernel is not intended for serious research.
    We strongly recommend against using it for any meaningful analyses.
    """

    def __init__(self, model_dir="auto"):
        # Use private attribute to hide value from get_params.
        # get_params will find model_dir=None, which is *very important* when a path is provided.
        self._model_dir = model_dir
        LGR.warning(
            "The Peaks2Maps kernel transformer is not intended for serious research. "
            "We strongly recommend against using it for any meaningful analyses."
        )

    def _transform(self, mask, coordinates):
        transformed = []
        coordinates_list = []
        ids = []
        for id_, data in coordinates.groupby("id"):
            ijk = np.vstack((data.i.values, data.j.values, data.k.values)).T.astype(int)
            xyz = vox2mm(ijk, mask.affine)
            coordinates_list.append(xyz)
            ids.append(id_)

        imgs = compute_p2m_ma(coordinates_list, skip_out_of_bounds=True, model_dir=self._model_dir)
        resampled_imgs = []
        for img in imgs:
            resampled_imgs.append(image.resample_to_img(img, mask).get_fdata())
        transformed = list(zip(resampled_imgs, ids))
        return transformed



"""Text extraction tools."""
import logging
import os.path as op

import pandas as pd
from sklearn.feature_extraction.text import CountVectorizer, TfidfVectorizer

from ..utils import get_resource_path

LGR = logging.getLogger(__name__)


def generate_counts(text_df, text_column="abstract", tfidf=True, min_df=50, max_df=0.5):
    """Generate tf-idf weights for unigrams/bigrams derived from textual data.

    Parameters
    ----------
    text_df : (D x 2) :obj:`pandas.DataFrame`
        A DataFrame with two columns ('id' and 'text'). D = document.

    Returns
    -------
    weights_df : (D x T) :obj:`pandas.DataFrame`
        A DataFrame where the index is 'id' and the columns are the
        unigrams/bigrams derived from the data. D = document. T = term.
    """
    if text_column not in text_df.columns:
        raise ValueError('Column "{0}" not found in DataFrame'.format(text_column))

    # Remove rows with empty text cells
    orig_ids = text_df["id"].tolist()
    text_df = text_df.fillna("")
    keep_ids = text_df.loc[text_df[text_column] != "", "id"]
    text_df = text_df.loc[text_df["id"].isin(keep_ids)]

    if len(keep_ids) != len(orig_ids):
        LGR.info("Retaining {0}/{1} studies".format(len(keep_ids), len(orig_ids)))

    ids = text_df["id"].tolist()
    text = text_df[text_column].tolist()
    stoplist = op.join(get_resource_path(), "neurosynth_stoplist.txt")
    with open(stoplist, "r") as fo:
        stop_words = fo.read().splitlines()

    if tfidf:
        vectorizer = TfidfVectorizer(
            min_df=min_df,
            max_df=max_df,
            ngram_range=(1, 2),
            vocabulary=None,
            stop_words=stop_words,
        )
    else:
        vectorizer = CountVectorizer(
            min_df=min_df,
            max_df=max_df,
            ngram_range=(1, 2),
            vocabulary=None,
            stop_words=stop_words,
        )
    weights = vectorizer.fit_transform(text).toarray()
    names = vectorizer.get_feature_names()
    names = [str(name) for name in names]
    weights_df = pd.DataFrame(weights, columns=names, index=ids)
    weights_df.index.name = "id"
    return weights_df



"""Coordinate-based meta-analytic estimators."""
from .ale import ALE, SCALE, ALESubtraction
from .mkda import KDA, MKDAChi2, MKDADensity

__all__ = ["ALE", "ALESubtraction", "SCALE", "MKDADensity", "MKDAChi2", "KDA"]



"""Test nimare.annotate.cogat (Cognitive Atlas extraction methods)."""
import os.path as op

import pandas as pd

import nimare
from nimare import annotate, extract

from .utils import get_test_data_path


def test_cogat():
    """A smoke test for CogAt-related functions."""
    # A small test dataset with abstracts
    ns_dset_laird = nimare.dataset.Dataset.load(
        op.join(get_test_data_path(), "neurosynth_laird_studies.pkl.gz")
    )
    cogat = extract.download_cognitive_atlas(data_dir=get_test_data_path(), overwrite=False)
    id_df = pd.read_csv(cogat["ids"])
    rel_df = pd.read_csv(cogat["relationships"])
    weights = {"isKindOf": 1, "isPartOf": 1, "inCategory": 1}
    counts_df, rep_text_df = annotate.cogat.extract_cogat(
        ns_dset_laird.texts, id_df, text_column="abstract"
    )
    assert "id" in ns_dset_laird.texts.columns
    expanded_df = annotate.cogat.expand_counts(counts_df, rel_df, weights)
    assert isinstance(expanded_df, pd.DataFrame)


def test_CogAtLemmatizer():
    """A smoke test for CogAtLemmatizer."""
    cogat = extract.download_cognitive_atlas(data_dir=get_test_data_path(), overwrite=False)
    id_df = pd.read_csv(cogat["ids"])
    id_df = id_df.loc[id_df["id"] == "trm_4aae62e4ad209"]
    lem = annotate.cogat.CogAtLemmatizer(id_df)
    true_text = "trm_4aae62e4ad209 is great"
    test_text = "Cognitive control is great"
    assert lem.transform(test_text) == true_text



"""Test nimare.dataset (Dataset IO/transformations)."""
import os.path as op

import nibabel as nib
import numpy as np

import nimare
from nimare import dataset
from nimare.tests.utils import get_test_data_path


def test_dataset_smoke():
    """Smoke test for nimare.dataset.Dataset initialization and get methods."""
    db_file = op.join(get_test_data_path(), "neurosynth_dset.json")
    dset = dataset.Dataset(db_file)
    dset.update_path(get_test_data_path())
    assert isinstance(dset, nimare.dataset.Dataset)
    methods = [dset.get_images, dset.get_labels, dset.get_metadata, dset.get_texts]
    for method in methods:
        assert isinstance(method(), list)
        assert isinstance(method(ids=dset.ids[:5]), list)
        assert isinstance(method(ids=dset.ids[0]), list)
    assert isinstance(dset.get_images(imtype="beta"), list)
    assert isinstance(dset.get_metadata(field="sample_sizes"), list)
    assert isinstance(dset.get_studies_by_label("cogat_cognitive_control"), list)
    assert isinstance(dset.get_studies_by_coordinate(np.array([[20, 20, 20]])), list)
    mask_data = np.zeros(dset.masker.mask_img.shape, int)
    mask_data[40, 40, 40] = 1
    mask_img = nib.Nifti1Image(mask_data, dset.masker.mask_img.affine)
    assert isinstance(dset.get_studies_by_mask(mask_img), list)


def test_empty_dset():
    """Smoke test for initialization with an empty Dataset."""
    # dictionary with no information
    minimal_dict = {"study-0": {"contrasts": {"1": {}}}}
    dataset.Dataset(minimal_dict)



"""Tests for the nimare.generate module."""
from contextlib import ExitStack as does_not_raise

import pytest
from numpy.random import RandomState

from ..dataset import Dataset
from ..generate import (
    _array_like,
    _create_foci,
    _create_source,
    create_coordinate_dataset,
    create_neurovault_dataset,
)


@pytest.mark.parametrize(
    "kwargs,expectation",
    [
        pytest.param(
            {
                "foci": [(0, 0, 0)],
                "foci_percentage": "60%",
                "fwhm": 10.0,
                "n_studies": 5,
                "n_noise_foci": 2,
                "rng": RandomState(seed=42),
                "space": "MNI",
            },
            does_not_raise(),
            id="specify_foci_coord",
        ),
        pytest.param(
            {
                "foci": 1,
                "foci_percentage": "60%",
                "fwhm": 10.0,
                "n_studies": 5,
                "n_noise_foci": 2,
                "rng": RandomState(seed=42),
                "space": "MNI",
            },
            does_not_raise(),
            id="integer_foci",
        ),
        pytest.param(
            {
                "foci": 0,
                "foci_percentage": "60%",
                "fwhm": 10.0,
                "n_studies": 5,
                "n_noise_foci": 0,
                "rng": RandomState(seed=42),
                "space": "MNI",
            },
            does_not_raise(),
            id="no_foci",
        ),
    ],
)
def test_create_foci(kwargs, expectation):
    """Smoke test for _create_foci."""
    with expectation:
        ground_truth_foci, foci_dict = _create_foci(**kwargs)
    if isinstance(expectation, does_not_raise):
        assert all(isinstance(key, int) for key in foci_dict)
        assert all(isinstance(coord, tuple) for coord in ground_truth_foci)


def test_create_source():
    """Smoke test for _create_source."""
    source_dict = _create_source(foci={0: [(0, 0, 0)]}, sample_sizes=[25])
    assert source_dict["study-0"]["contrasts"]["1"]["metadata"]["sample_sizes"] == [25]


@pytest.mark.parametrize(
    "kwargs,expectation",
    [
        pytest.param(
            {
                "foci": 2,
                "foci_percentage": 1.0,
                "fwhm": 10.0,
                "sample_size": (10, 20),
                "n_studies": 5,
                "n_noise_foci": 0,
                "seed": 42,
                "space": "MNI",
            },
            does_not_raise(),
            id="random_sample_size",
        ),
        pytest.param(
            {
                "foci": [(0, 0, 0), (0, 10, 10)],
                "foci_percentage": "100%",
                "fwhm": 10.0,
                "sample_size": [30] * 5,
                "n_studies": 5,
                "n_noise_foci": 0,
                "seed": 42,
                "space": "MNI",
            },
            does_not_raise(),
            id="specified_sample_size",
        ),
        pytest.param(
            {
                "foci": 2,
                "fwhm": 10.0,
                "sample_size": [30] * 4,
                "n_studies": 5,
                "n_noise_foci": 0,
                "seed": 42,
                "space": "MNI",
            },
            pytest.raises(ValueError),
            id="incorrect_sample_size_list",
        ),
        pytest.param(
            {
                "foci": 0,
                "foci_percentage": 1.0,
                "fwhm": 10.0,
                "sample_size": (10, 20),
                "n_studies": 5,
                "n_noise_foci": 0,
                "seed": 42,
                "space": "MNI",
            },
            does_not_raise(),
            id="no_foci",
        ),
        pytest.param(
            {
                "foci": 0,
                "foci_percentage": "50%",
                "fwhm": 10.0,
                "sample_size": (10, 20),
                "n_studies": 5,
                "n_noise_foci": 10,
                "seed": 42,
                "space": "MNI",
            },
            does_not_raise(),
            id="only_noise_foci",
        ),
        pytest.param(
            {
                "foci": 1,
                "foci_percentage": "50%",
                "fwhm": 10.0,
                "sample_size": (10, 20),
                "n_studies": 5,
                "n_noise_foci": 0,
                "seed": 42,
                "space": "MNI",
            },
            does_not_raise(),
            id="insufficient_foci",
        ),
        pytest.param(
            {
                "foci": "INVALID_FOCI",
                "foci_percentage": "50%",
                "fwhm": 10.0,
                "sample_size": (10, 20),
                "n_studies": 5,
                "n_noise_foci": 0,
                "seed": 42,
                "space": "MNI",
            },
            pytest.raises(ValueError),
            id="invalid_foci",
        ),
        pytest.param(
            {
                "foci": 1,
                "foci_percentage": "INVALID_PERCENT",
                "fwhm": 10.0,
                "sample_size": (10, 20),
                "n_studies": 5,
                "n_noise_foci": 0,
                "seed": 42,
                "space": "MNI",
            },
            pytest.raises(ValueError),
            id="invalid_percent",
        ),
        pytest.param(
            {
                "foci": 1,
                "foci_percentage": "60%",
                "fwhm": 10.0,
                "sample_size": "INVALID_SAMPLE_SIZE",
                "n_studies": 5,
                "n_noise_foci": 0,
                "seed": 42,
                "space": "MNI",
            },
            pytest.raises(ValueError),
            id="invalid_sample_size",
        ),
        pytest.param(
            {
                "foci": 1,
                "foci_percentage": "60%",
                "fwhm": 10.0,
                "sample_size": 30,
                "n_studies": 5,
                "n_noise_foci": 0,
                "seed": 42,
                "space": "INVALID_SPACE",
            },
            pytest.raises(NotImplementedError),
            id="invalid_space",
        ),
    ],
)
def test_create_coordinate_dataset(kwargs, expectation):
    """Create a coordinate Dataset according to parameters."""
    with expectation:
        ground_truth_foci, dataset = create_coordinate_dataset(**kwargs)
    if isinstance(expectation, does_not_raise):
        assert isinstance(dataset, Dataset)
        assert len(dataset.ids) == kwargs["n_studies"]
        # test if the number of observed coordinates in the dataset is correct
        if _array_like(kwargs["foci"]):
            n_foci = len(kwargs["foci"])
        else:
            n_foci = kwargs["foci"]
        expected_coordinate_number = max(
            kwargs["n_studies"],
            (kwargs["n_studies"] * n_foci) + (kwargs["n_studies"] * kwargs["n_noise_foci"]),
        )
        assert len(dataset.coordinates) == expected_coordinate_number


def test_create_neurovault_dataset():
    """Test creating a neurovault dataset."""
    dset = create_neurovault_dataset(
        collection_ids=(8836,),
        contrasts={"animal": "as-Animal"},
    )
    expected_columns = {"beta", "t", "varcope", "z"}
    assert expected_columns.issubset(dset.images.columns)



"""Test nimare.decode.discrete.

Tests for nimare.decode.discrete.gclda_decode_roi are in test_annotate_gclda.
"""
import pandas as pd
import pytest

from nimare.decode import discrete


def test_neurosynth_decode(testdata_laird):
    """Smoke test for discrete.neurosynth_decode."""
    ids = testdata_laird.ids[:5]
    features = testdata_laird.annotations.columns.tolist()[5:10]
    decoded_df = discrete.neurosynth_decode(
        testdata_laird.coordinates,
        testdata_laird.annotations,
        ids=ids,
        features=features,
        correction=None,
    )
    assert isinstance(decoded_df, pd.DataFrame)


def test_brainmap_decode(testdata_laird):
    """Smoke test for discrete.brainmap_decode."""
    ids = testdata_laird.ids[:5]
    features = testdata_laird.annotations.columns.tolist()[5:10]
    decoded_df = discrete.brainmap_decode(
        testdata_laird.coordinates,
        testdata_laird.annotations,
        ids=ids,
        features=features,
        correction=None,
    )
    assert isinstance(decoded_df, pd.DataFrame)


def test_NeurosynthDecoder(testdata_laird):
    """Smoke test for discrete.NeurosynthDecoder."""
    ids = testdata_laird.ids[:5]
    decoder = discrete.NeurosynthDecoder()
    decoder.fit(testdata_laird)
    decoded_df = decoder.transform(ids=ids)
    assert isinstance(decoded_df, pd.DataFrame)


def test_NeurosynthDecoder_featuregroup(testdata_laird):
    """Smoke test for discrete.NeurosynthDecoder with feature group selection."""
    ids = testdata_laird.ids[:5]
    decoder = discrete.NeurosynthDecoder(feature_group="Neurosynth_TFIDF")
    decoder.fit(testdata_laird)
    decoded_df = decoder.transform(ids=ids)
    assert isinstance(decoded_df, pd.DataFrame)


def test_NeurosynthDecoder_featuregroup_failure(testdata_laird):
    """Smoke test for NeurosynthDecoder with feature group selection and no detected features."""
    decoder = discrete.NeurosynthDecoder(feature_group="Neurosynth_TFIDF", features=["01", "05"])
    with pytest.raises(Exception):
        decoder.fit(testdata_laird)


def test_BrainMapDecoder(testdata_laird):
    """Smoke test for discrete.BrainMapDecoder."""
    ids = testdata_laird.ids[:5]
    decoder = discrete.BrainMapDecoder()
    decoder.fit(testdata_laird)
    decoded_df = decoder.transform(ids=ids)
    assert isinstance(decoded_df, pd.DataFrame)


def test_BrainMapDecoder_failure(testdata_laird):
    """Smoke test for discrete.BrainMapDecoder where there are no features left."""
    decoder = discrete.BrainMapDecoder(features=["doggy"])
    with pytest.raises(Exception):
        decoder.fit(testdata_laird)



"""CBMA methods from the ALE and MKDA families."""
import logging
import multiprocessing as mp

import nibabel as nib
import numpy as np
import pandas as pd
from scipy import ndimage
from tqdm.auto import tqdm

from ...base import MetaEstimator
from ...results import MetaResult
from ...stats import null_to_p, nullhist_to_p
from ...transforms import p_to_z
from ...utils import add_metadata_to_dataframe, check_type, use_memmap
from ..kernel import KernelTransformer

LGR = logging.getLogger(__name__)


class CBMAEstimator(MetaEstimator):
    """Base class for coordinate-based meta-analysis methods.

    Parameters
    ----------
    kernel_transformer : :obj:`nimare.meta.kernel.KernelTransformer`, optional
        Kernel with which to convolve coordinates from dataset. Default is
        ALEKernel.
    *args
        Optional arguments to the :obj:`nimare.base.MetaEstimator` __init__
        (called automatically).
    **kwargs
        Optional keyword arguments to the :obj:`nimare.base.MetaEstimator`
        __init__ (called automatically).
    """

    _required_inputs = {"coordinates": ("coordinates", None)}

    def __init__(self, kernel_transformer, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Get kernel transformer
        kernel_args = {
            k.split("kernel__")[1]: v for k, v in kwargs.items() if k.startswith("kernel__")
        }
        kernel_transformer = check_type(kernel_transformer, KernelTransformer, **kernel_args)
        self.kernel_transformer = kernel_transformer

    @use_memmap(LGR, n_files=1)
    def _fit(self, dataset):
        """
        Perform coordinate-based meta-analysis on dataset.

        Parameters
        ----------
        dataset : :obj:`nimare.dataset.Dataset`
            Dataset to analyze.
        """
        self.dataset = dataset
        self.masker = self.masker or dataset.masker
        self.null_distributions_ = {}

        ma_values = self._collect_ma_maps(
            coords_key="coordinates",
            maps_key="ma_maps",
            fname_idx=0,
        )

        self.weight_vec_ = self._compute_weights(ma_values)

        stat_values = self.compute_summarystat(ma_values)

        # Determine null distributions for summary stat (OF) to p conversion
        self._determine_histogram_bins(ma_values)
        if self.null_method.startswith("approximate"):
            self._compute_null_approximate(ma_values)

        elif self.null_method == "montecarlo":
            self._compute_null_montecarlo(n_iters=self.n_iters, n_cores=self.n_cores)

        else:
            # A hidden option only used for internal validation/testing
            self._compute_null_reduced_montecarlo(ma_values, n_iters=self.n_iters)

        p_values, z_values = self._summarystat_to_p(stat_values, null_method=self.null_method)

        images = {"stat": stat_values, "p": p_values, "z": z_values}
        return images

    def _compute_weights(self, ma_values):
        """Perform optional weight computation routine.

        Takes an array of meta-analysis values as input and returns an array
        of the same shape, weighted as desired.
        Can be ignored by algorithms that don't support weighting.
        """
        return None

    def _preprocess_input(self, dataset):
        """Mask required input images using either the dataset's mask or the estimator's.

        Also, insert required metadata into coordinates DataFrame.
        """
        super()._preprocess_input(dataset)

        # All extra (non-ijk) parameters for a kernel should be overrideable as
        # parameters to __init__, so we can access them with get_params()
        kt_args = list(self.kernel_transformer.get_params().keys())

        # Integrate "sample_size" from metadata into DataFrame so that
        # kernel_transformer can access it.
        if "sample_size" in kt_args:
            self.inputs_["coordinates"] = add_metadata_to_dataframe(
                dataset,
                self.inputs_["coordinates"],
                metadata_field="sample_sizes",
                target_column="sample_size",
                filter_func=np.mean,
            )

    def _collect_ma_maps(self, coords_key="coordinates", maps_key="ma_maps", fname_idx=0):
        """Collect modeled activation maps from Estimator inputs.

        Parameters
        ----------
        coords_key : :obj:`str`, optional
            Key to Estimator.inputs_ dictionary containing coordinates DataFrame.
            This key should **always** be present.
        maps_key : :obj:`str`, optional
            Key to Estimator.inputs_ dictionary containing list of MA map files.
            This key should only be present if the kernel transformer was already fitted to the
            input Dataset.
        fname_idx : :obj:`int`, optional
            When the Estimator is set with ``low_memory = True``, there is a ``memmap_filenames``
            attribute that is a list of filenames or Nones. This parameter specifies which item
            in that list should be used for a memory-mapped array.

        Returns
        -------
        ma_maps : :obj:`numpy.ndarray`
            2D numpy array of shape (n_studies, n_voxels) with MA values.
        """
        if maps_key in self.inputs_.keys():
            LGR.debug(f"Loading pre-generated MA maps ({maps_key}).")
            if self.low_memory:
                temp = self.masker.transform(self.inputs_[maps_key][0])
                unmasked_shape = (len(self.inputs_[maps_key]), temp.size)
                ma_maps = np.memmap(
                    self.memmap_filenames[fname_idx],
                    dtype=temp.dtype,
                    mode="w+",
                    shape=unmasked_shape,
                )
                for i, f in enumerate(self.inputs_[maps_key]):
                    ma_maps[i, :] = self.masker.transform(f)
            else:
                ma_maps = self.masker.transform(self.inputs_[maps_key])
        else:
            LGR.debug(f"Generating MA maps from coordinates ({coords_key}).")
            ma_maps = self.kernel_transformer.transform(
                self.inputs_[coords_key],
                masker=self.masker,
                return_type="array",
            )
        return ma_maps

    def compute_summarystat(self, data):
        """Compute OF scores from data.

        Parameters
        ----------
        data : array, pandas.DataFrame, or list of img_like
            Data from which to estimate ALE scores.
            The data can be:
            (1) a 1d contrast-len or 2d contrast-by-voxel array of MA values,
            (2) a DataFrame containing coordinates to produce MA values,
            or (3) a list of imgs containing MA values.

        Returns
        -------
        stat_values : 1d array
            OF values. One value per voxel.
        """
        if isinstance(data, pd.DataFrame):
            ma_values = self.kernel_transformer.transform(
                data, masker=self.masker, return_type="array"
            )
        elif isinstance(data, list):
            ma_values = self.masker.transform(data)
        elif isinstance(data, np.ndarray):
            ma_values = data
        elif not isinstance(data, np.ndarray):
            raise ValueError('Unsupported data type "{}"'.format(type(data)))

        # Apply weights before returning
        return self._compute_summarystat(ma_values)

    def _compute_summarystat(self, ma_values):
        """Compute summary statistic according to estimator-specific method.

        Must be overriden by subclasses.
        Input and output are both numpy arrays; the output must
        aggregate over the 0th dimension of the input.
        (i.e., if the input has K dimensions, the output has K - 1 dimensions.)
        """
        pass

    def _summarystat_to_p(self, stat_values, null_method="approximate"):
        """Compute p- and z-values from summary statistics (e.g., ALE scores).

        Uses either histograms from approximate null or null distribution from montecarlo null.

        Parameters
        ----------
        stat_values : 1D array_like
            Array of summary statistic values from estimator.
        null_method : {"approximate", "montecarlo"}, optional
            Whether to use approximate null or montecarlo null.
            Default is "approximate".

        Returns
        -------
        p_values, z_values : 1D array
            P- and Z-values for statistic values.
            Same shape as stat_values.
        """
        if null_method.startswith("approximate"):
            assert "histogram_bins" in self.null_distributions_.keys()
            assert "histweights_corr-none_method-approximate" in self.null_distributions_.keys()

            p_values = nullhist_to_p(
                stat_values,
                self.null_distributions_["histweights_corr-none_method-approximate"],
                self.null_distributions_["histogram_bins"],
            )

        elif null_method == "montecarlo":
            assert "histogram_bins" in self.null_distributions_.keys()
            assert "histweights_corr-none_method-montecarlo" in self.null_distributions_.keys()
            p_values = nullhist_to_p(
                stat_values,
                self.null_distributions_["histweights_corr-none_method-montecarlo"],
                self.null_distributions_["histogram_bins"],
            )

        elif null_method == "reduced_montecarlo":
            assert "values_corr-none_method-reducedMontecarlo" in self.null_distributions_.keys()
            p_values = null_to_p(
                stat_values,
                self.null_distributions_["values_corr-none_method-reducedMontecarlo"],
                tail="upper",
            )

        else:
            raise ValueError("Argument 'null_method' must be one of: 'approximate', 'montecarlo'.")

        z_values = p_to_z(p_values, tail="one")
        return p_values, z_values

    def _p_to_summarystat(self, p, null_method=None):
        """Compute a summary statistic threshold that corresponds to the provided p-value.

        Uses either histograms from approximate null or null distribution from montecarlo null.

        Parameters
        ----------
        p : The p-value that corresponds to the summary statistic threshold
        null_method : {None, "approximate", "montecarlo"}, optional
            Whether to use approximate null or montecarlo null. If None, defaults to using
            whichever method was set at initialization.

        Returns
        -------
        ss : float
            A float giving the summary statistic value corresponding to the passed p.
        """
        if null_method is None:
            null_method = self.null_method

        if null_method.startswith("approximate"):
            assert "histogram_bins" in self.null_distributions_.keys()
            assert "histweights_corr-none_method-approximate" in self.null_distributions_.keys()

            # Convert unnormalized histogram weights to null distribution
            histogram_weights = self.null_distributions_[
                "histweights_corr-none_method-approximate"
            ]
            null_distribution = histogram_weights / np.sum(histogram_weights)
            null_distribution = np.cumsum(null_distribution[::-1])[::-1]
            null_distribution /= np.max(null_distribution)
            null_distribution = np.squeeze(null_distribution)

            # Desired bin is the first one _before_ the target p-value (for consistency
            # with the montecarlo null).
            ss_idx = np.maximum(0, np.where(null_distribution <= p)[0][0] - 1)
            ss = self.null_distributions_["histogram_bins"][ss_idx]

        elif null_method == "montecarlo":
            assert "histogram_bins" in self.null_distributions_.keys()
            assert "histweights_corr-none_method-montecarlo" in self.null_distributions_.keys()

            hist_weights = self.null_distributions_["histweights_corr-none_method-montecarlo"]
            # Desired bin is the first one _before_ the target p-value (for consistency
            # with the montecarlo null).
            ss_idx = np.maximum(0, np.where(hist_weights <= p)[0][0] - 1)
            ss = self.null_distributions_["histogram_bins"][ss_idx]

        elif null_method == "reduced_montecarlo":
            assert "values_corr-none_method-reducedMontecarlo" in self.null_distributions_.keys()
            null_dist = np.sort(
                self.null_distributions_["values_corr-none_method-reducedMontecarlo"]
            )
            n_vals = len(null_dist)
            ss_idx = np.floor(p * n_vals).astype(int)
            ss = null_dist[-ss_idx]

        else:
            raise ValueError("Argument 'null_method' must be one of: 'approximate', 'montecarlo'.")

        return ss

    def _compute_null_reduced_montecarlo(self, ma_maps, n_iters=10000):
        """Compute uncorrected null distribution using the reduced montecarlo method.

        This method is much faster than the full montecarlo approach, but is still slower than the
        approximate method. Given that its resolution is roughly the same as the approximate
        method, we recommend against using this method.

        Parameters
        ----------
        ma_maps : (C x V) array
            Contrast by voxel array of MA values, after weighting with
            weight_vec.

        Notes
        -----
        This method adds one entry to the null_distributions_ dict attribute:
        "values_corr-none_method-reducedMontecarlo".
        """
        n_studies, n_voxels = ma_maps.shape
        null_ijk = np.random.choice(np.arange(n_voxels), (n_iters, n_studies))
        iter_ma_values = ma_maps[np.arange(n_studies), tuple(null_ijk)].T
        null_dist = self.compute_summarystat(iter_ma_values)
        self.null_distributions_["values_corr-none_method-reducedMontecarlo"] = null_dist

    def _compute_null_montecarlo_permutation(self, params):
        """Run a single Monte Carlo permutation of a dataset.

        Does the shared work between uncorrected stat-to-p conversion and vFWE.

        Parameters
        ----------
        params : tuple
            A tuple containing 2 elements, respectively providing (1) the permuted
            coordinates and (2) the original coordinate DataFrame.

        Returns
        -------
        counts : 1D array_like
            Weights associated with the attribute `null_distributions_["histogram_bins"]`.
        """
        iter_ijk, iter_df = params

        iter_ijk = np.squeeze(iter_ijk)
        iter_df[["i", "j", "k"]] = iter_ijk

        iter_ma_maps = self.kernel_transformer.transform(
            iter_df, masker=self.masker, return_type="array"
        )
        iter_ss_map = self.compute_summarystat(iter_ma_maps)

        # Get bin edges for histogram
        bin_centers = self.null_distributions_["histogram_bins"]
        step_size = bin_centers[1] - bin_centers[0]
        bin_edges = bin_centers - (step_size / 2)
        bin_edges = np.append(bin_centers, bin_centers[-1] + step_size)

        counts, _ = np.histogram(iter_ss_map, bins=bin_edges, density=False)
        return counts

    def _compute_null_montecarlo(self, n_iters, n_cores):
        """Compute uncorrected null distribution using Monte Carlo method.

        Parameters
        ----------
        n_iters : int
            Number of permutations.
        n_cores : int
            Number of cores to use.

        Notes
        -----
        This method adds two entries to the null_distributions_ dict attribute:
        "histweights_corr-none_method-montecarlo" and
        "histweights_level-voxel_corr-fwe_method-montecarlo".
        """
        null_ijk = np.vstack(np.where(self.masker.mask_img.get_fdata())).T

        n_cores = self._check_ncores(n_cores)

        rand_idx = np.random.choice(
            null_ijk.shape[0], size=(self.inputs_["coordinates"].shape[0], n_iters)
        )
        rand_ijk = null_ijk[rand_idx, :]
        iter_ijks = np.split(rand_ijk, rand_ijk.shape[1], axis=1)
        iter_df = self.inputs_["coordinates"].copy()
        iter_dfs = [iter_df] * n_iters

        params = zip(iter_ijks, iter_dfs)
        if n_cores == 1:
            perm_histograms = []
            for pp in tqdm(params, total=n_iters):
                perm_histograms.append(self._compute_null_montecarlo_permutation(pp))

        else:
            with mp.Pool(n_cores) as p:
                perm_histograms = list(
                    tqdm(p.imap(self._compute_null_montecarlo_permutation, params), total=n_iters)
                )

        perm_histograms = np.vstack(perm_histograms)
        self.null_distributions_["histweights_corr-none_method-montecarlo"] = np.sum(
            perm_histograms, axis=0
        )

        def get_last_bin(arr1d):
            """Index the last location in a 1D array with a non-zero value."""
            if np.any(arr1d):
                last_bin = np.where(arr1d)[0][-1]
            else:
                last_bin = 0
            return last_bin

        fwe_voxel_max = np.apply_along_axis(get_last_bin, 1, perm_histograms)
        self.null_distributions_[
            "histweights_level-voxel_corr-fwe_method-montecarlo"
        ] = fwe_voxel_max

    def _correct_fwe_montecarlo_permutation(self, params):
        """Run a single Monte Carlo permutation of a dataset.

        Does the shared work between vFWE and cFWE.

        Parameters
        ----------
        params : tuple
            A tuple containing 4 elements, respectively providing (1) the permuted
            coordinates; (2) the original coordinate DataFrame; (3) a 3d structuring
            array passed to ndimage.label; and (4) the voxel-wise intensity threshold.

        Returns
        -------
        (iter_max value, iter_max_cluster)
            A 2-tuple of floats giving the maximum voxel-wise value, and maximum
            cluster size for the permuted dataset.
        """
        iter_ijk, iter_df, conn, voxel_thresh = params

        iter_ijk = np.squeeze(iter_ijk)
        iter_df[["i", "j", "k"]] = iter_ijk

        iter_ma_maps = self.kernel_transformer.transform(
            iter_df, masker=self.masker, return_type="array"
        )
        iter_ss_map = self.compute_summarystat(iter_ma_maps)
        iter_max_value = np.max(iter_ss_map)
        iter_ss_map = self.masker.inverse_transform(iter_ss_map).get_fdata().copy()
        iter_ss_map[iter_ss_map <= voxel_thresh] = 0

        labeled_matrix = ndimage.measurements.label(iter_ss_map, conn)[0]
        _, clust_sizes = np.unique(labeled_matrix, return_counts=True)
        clust_sizes = clust_sizes[1:]  # First cluster is zeros in matrix
        if clust_sizes.size:
            iter_max_cluster = np.max(clust_sizes)
        else:
            iter_max_cluster = 0
        return iter_max_value, iter_max_cluster

    def correct_fwe_montecarlo(
        self, result, voxel_thresh=0.001, n_iters=10000, n_cores=-1, vfwe_only=False
    ):
        """Perform FWE correction using the max-value permutation method.

        Only call this method from within a Corrector.

        Parameters
        ----------
        result : :obj:`nimare.results.MetaResult`
            Result object from a KDA meta-analysis.
        voxel_thresh : :obj:`float`, optional
            Cluster-defining p-value threshold. Default is 0.001.
        n_iters : :obj:`int`, optional
            Number of iterations to build the vFWE and cFWE null distributions.
            Default is 10000.
        n_cores : :obj:`int`, optional
            Number of cores to use for parallelization.
            If <=0, defaults to using all available cores. Default is -1.

        Returns
        -------
        images : :obj:`dict`
            Dictionary of 1D arrays corresponding to masked images generated by
            the correction procedure. The following arrays are generated by
            this method: 'vthresh', 'logp_level-cluster', and 'logp_level-voxel'.

        See Also
        --------
        nimare.correct.FWECorrector : The Corrector from which to call this method.

        Examples
        --------
        >>> meta = MKDADensity()
        >>> result = meta.fit(dset)
        >>> corrector = FWECorrector(method='montecarlo', voxel_thresh=0.01,
                                     n_iters=5, n_cores=1)
        >>> cresult = corrector.transform(result)
        """
        stat_values = result.get_map("stat", return_type="array")
        if vfwe_only:
            assert self.null_method == "montecarlo"

            LGR.info("Using precalculated histogram for voxel-level FWE correction.")

            # Determine p- and z-values from stat values and null distribution.
            p_vfwe_values = nullhist_to_p(
                stat_values,
                self.null_distributions_["histweights_level-voxel_corr-fwe_method-montecarlo"],
                self.null_distributions_["histogram_bins"],
            )

        else:
            null_ijk = np.vstack(np.where(self.masker.mask_img.get_fdata())).T

            n_cores = self._check_ncores(n_cores)

            # Identify summary statistic corresponding to intensity threshold
            ss_thresh = self._p_to_summarystat(voxel_thresh)

            rand_idx = np.random.choice(
                null_ijk.shape[0], size=(self.inputs_["coordinates"].shape[0], n_iters)
            )
            rand_ijk = null_ijk[rand_idx, :]
            iter_ijks = np.split(rand_ijk, rand_ijk.shape[1], axis=1)
            iter_df = self.inputs_["coordinates"].copy()
            iter_dfs = [iter_df] * n_iters

            # Find number of voxels per cluster (includes 0, which is empty space in
            # the matrix)
            conn = np.zeros((3, 3, 3), int)
            conn[:, :, 1] = 1
            conn[:, 1, :] = 1
            conn[1, :, :] = 1

            # Define parameters
            iter_conn = [conn] * n_iters
            iter_ss_thresh = [ss_thresh] * n_iters
            params = zip(iter_ijks, iter_dfs, iter_conn, iter_ss_thresh)

            if n_cores == 1:
                perm_results = []
                for pp in tqdm(params, total=n_iters):
                    perm_results.append(self._correct_fwe_montecarlo_permutation(pp))

            else:
                with mp.Pool(n_cores) as p:
                    perm_results = list(
                        tqdm(
                            p.imap(self._correct_fwe_montecarlo_permutation, params), total=n_iters
                        )
                    )

            fwe_voxel_max, fwe_clust_max = zip(*perm_results)

            # Cluster-level FWE
            thresh_stat_values = self.masker.inverse_transform(stat_values).get_fdata()
            thresh_stat_values[thresh_stat_values <= ss_thresh] = 0
            labeled_matrix, _ = ndimage.measurements.label(thresh_stat_values, conn)

            _, idx, sizes = np.unique(labeled_matrix, return_inverse=True, return_counts=True)
            # first cluster has value 0 (i.e., all non-zero voxels in brain), so replace
            # with 0, which gives us a p-value of 1.
            sizes[0] = 0
            p_vals = null_to_p(sizes, fwe_clust_max, "upper")
            p_cfwe_map = p_vals[np.reshape(idx, labeled_matrix.shape)]

            p_cfwe_values = np.squeeze(
                self.masker.transform(nib.Nifti1Image(p_cfwe_map, self.masker.mask_img.affine))
            )
            logp_cfwe_values = -np.log10(p_cfwe_values)
            logp_cfwe_values[np.isinf(logp_cfwe_values)] = -np.log10(np.finfo(float).eps)
            z_cfwe_values = p_to_z(p_cfwe_values, tail="one")

            # Voxel-level FWE
            LGR.info("Using null distribution for voxel-level FWE correction.")
            p_vfwe_values = null_to_p(stat_values, fwe_voxel_max, tail="upper")
            self.null_distributions_[
                "values_level-voxel_corr-fwe_method-montecarlo"
            ] = fwe_voxel_max
            self.null_distributions_[
                "values_level-cluster_corr-fwe_method-montecarlo"
            ] = fwe_clust_max

        z_vfwe_values = p_to_z(p_vfwe_values, tail="one")
        logp_vfwe_values = -np.log10(p_vfwe_values)
        logp_vfwe_values[np.isinf(logp_vfwe_values)] = -np.log10(np.finfo(float).eps)

        if vfwe_only:
            # Write out unthresholded value images
            images = {
                "logp_level-voxel": logp_vfwe_values,
                "z_level-voxel": z_vfwe_values,
            }

        else:
            # Write out unthresholded value images
            images = {
                "logp_level-voxel": logp_vfwe_values,
                "z_level-voxel": z_vfwe_values,
                "logp_level-cluster": logp_cfwe_values,
                "z_level-cluster": z_cfwe_values,
            }

        return images


class PairwiseCBMAEstimator(CBMAEstimator):
    """Base class for pairwise coordinate-based meta-analysis methods.

    Parameters
    ----------
    kernel_transformer : :obj:`nimare.meta.kernel.KernelTransformer`, optional
        Kernel with which to convolve coordinates from dataset. Default is
        ALEKernel.
    *args
        Optional arguments to the :obj:`nimare.base.MetaEstimator` __init__
        (called automatically).
    **kwargs
        Optional keyword arguments to the :obj:`nimare.base.MetaEstimator`
        __init__ (called automatically).
    """

    def fit(self, dataset1, dataset2):
        """
        Fit Estimator to two Datasets.

        Parameters
        ----------
        dataset1/dataset2 : :obj:`nimare.dataset.Dataset`
            Dataset objects to analyze.

        Returns
        -------
        :obj:`nimare.results.MetaResult`
            Results of Estimator fitting.

        Notes
        -----
        The `fit` method is a light wrapper that runs input validation and
        preprocessing before fitting the actual model. Estimators' individual
        "fitting" methods are implemented as `_fit`, although users should
        call `fit`.
        """
        self._validate_input(dataset1)
        self._validate_input(dataset2)

        # grab and override
        self._preprocess_input(dataset1)
        if "ma_maps" in self.inputs_.keys():
            # Grab pre-generated MA maps
            self.inputs_["ma_maps1"] = self.inputs_.pop("ma_maps")

        self.inputs_["coordinates1"] = self.inputs_.pop("coordinates")

        # grab and override
        self._preprocess_input(dataset2)
        if "ma_maps" in self.inputs_.keys():
            # Grab pre-generated MA maps
            self.inputs_["ma_maps2"] = self.inputs_.pop("ma_maps")

        self.inputs_["coordinates2"] = self.inputs_.pop("coordinates")

        maps = self._fit(dataset1, dataset2)

        if hasattr(self, "masker") and self.masker is not None:
            masker = self.masker
        else:
            masker = dataset1.masker
        self.results = MetaResult(self, masker, maps)
        return self.results



"""Topic modeling with latent Dirichlet allocation via MALLET."""
import logging
import os
import os.path as op
import shutil
import subprocess

import numpy as np
import pandas as pd

from .. import references
from ..base import NiMAREBase
from ..due import due
from ..extract import download_mallet, utils

LGR = logging.getLogger(__name__)


@due.dcite(references.LDA, description="Introduces LDA.")
@due.dcite(references.MALLET, description="Citation for MALLET toolbox")
@due.dcite(
    references.LDAMODEL,
    description="First use of LDA for automated annotation of neuroimaging literature.",
)
class LDAModel(NiMAREBase):
    """
    Perform topic modeling using Latent Dirichlet Allocation (LDA).

    Build an LDA [1]_ topic model with the Java toolbox MALLET [2]_, as
    performed in [3]_.

    Parameters
    ----------
    text_df : :obj:`pandas.DataFrame`
        A pandas DataFrame with two columns ('id' and text_column) containing
        article text.
    text_column : :obj:`str`, optional
        Name of column in text_df that contains text. Default is 'abstract'.
    n_topics : :obj:`int`, optional
        Number of topics to generate. Default=50.
    n_iters : :obj:`int`, optional
        Number of iterations to run in training topic model. Default=1000.
    alpha : :obj:`float` or 'auto', optional
        The Dirichlet prior on the per-document topic distributions.
        Default: auto, which calculates 50 / n_topics, based on Poldrack et al.
        (2012).
    beta : :obj:`float`, optional
        The Dirichlet prior on the per-topic word distribution. Default: 0.001,
        based on Poldrack et al. (2012).

    Attributes
    ----------
    commands_ : :obj:`list` of :obj:`str`
        List of MALLET commands called to fit model.

    References
    ----------
    .. [1] Blei, David M., Andrew Y. Ng, and Michael I. Jordan. "Latent
        dirichlet allocation." Journal of machine Learning research 3.Jan
        (2003): 993-1022.
    .. [2] McCallum, Andrew Kachites. "Mallet: A machine learning for language
        toolkit." (2002).
    .. [3] Poldrack, Russell A., et al. "Discovering relations between mind,
        brain, and mental disorders using topic mapping." PLoS computational
        biology 8.10 (2012): e1002707.
        https://doi.org/10.1371/journal.pcbi.1002707
    """

    def __init__(
        self, text_df, text_column="abstract", n_topics=50, n_iters=1000, alpha="auto", beta=0.001
    ):
        mallet_dir = download_mallet()
        mallet_bin = op.join(mallet_dir, "bin/mallet")

        model_dir = utils._get_dataset_dir("mallet_model")
        text_dir = op.join(model_dir, "texts")

        if not op.isdir(model_dir):
            os.mkdir(model_dir)

        if alpha == "auto":
            alpha = 50.0 / n_topics
        elif not isinstance(alpha, float):
            raise ValueError('Argument alpha must be float or "auto"')

        self.params = {"n_topics": n_topics, "n_iters": n_iters, "alpha": alpha, "beta": beta}
        self.model_dir = model_dir

        # Check for presence of text files and convert if necessary
        if not op.isdir(text_dir):
            LGR.info("Texts folder not found. Creating text files...")
            os.mkdir(text_dir)

            # Remove rows with empty text cells
            orig_ids = text_df["id"].tolist()
            text_df = text_df.dropna(subset=[text_column])
            keep_ids = text_df["id"].tolist()

            if len(keep_ids) != len(orig_ids):
                LGR.info("Retaining {0}/{1} studies".format(len(keep_ids), len(orig_ids)))

            for id_ in text_df["id"].values:
                text = text_df.loc[text_df["id"] == id_, text_column].values[0]
                with open(op.join(text_dir, str(id_) + ".txt"), "w") as fo:
                    fo.write(text)

        # Run MALLET topic modeling
        LGR.info("Generating topics...")
        import_str = (
            "{mallet} import-dir "
            "--input {text_dir} "
            "--output {outdir}/topic-input.mallet "
            "--keep-sequence "
            "--remove-stopwords"
        ).format(mallet=mallet_bin, text_dir=text_dir, outdir=model_dir)

        train_str = (
            "{mallet} train-topics "
            "--input {out}/topic-input.mallet "
            "--num-topics {n_topics} "
            "--output-doc-topics {out}/doc_topics.txt "
            "--topic-word-weights-file {out}/topic_word_weights.txt "
            "--num-iterations {n_iters} "
            "--output-model {out}/saved_model.mallet "
            "--random-seed 1 "
            "--alpha {alpha} "
            "--beta {beta}"
        ).format(
            mallet=mallet_bin,
            out=model_dir,
            n_topics=self.params["n_topics"],
            n_iters=self.params["n_iters"],
            alpha=self.params["alpha"],
            beta=self.params["beta"],
        )
        self.commands_ = [import_str, train_str]

    def fit(self):
        """
        Fit LDA model to corpus.

        Attributes
        ----------
        p_topic_g_doc_ : :obj:`numpy.ndarray`
            Probability of each topic given a document
        p_word_g_topic_ : :obj:`numpy.ndarray`
            Probability of each word given a topic
        """
        subprocess.call(self.commands_[0], shell=True)
        subprocess.call(self.commands_[1], shell=True)

        # Read in and convert doc_topics and topic_keys.
        topic_names = ["topic_{0:03d}".format(i) for i in range(self.params["n_topics"])]

        # doc_topics: Topic weights for each paper.
        # The conversion here is pretty ugly at the moment.
        # First row should be dropped. First column is row number and can be used
        # as the index.
        # Second column is 'file: /full/path/to/id.txt' <-- Parse to get id.
        # After that, odd columns are topic numbers and even columns are the
        # weights for the topics in the preceding column. These columns are sorted
        # on an individual id basis by the weights.
        n_cols = (2 * self.params["n_topics"]) + 1
        dt_df = pd.read_csv(
            op.join(self.model_dir, "doc_topics.txt"),
            delimiter="\t",
            skiprows=1,
            header=None,
            index_col=0,
        )
        dt_df = dt_df[dt_df.columns[:n_cols]]

        # Get ids from filenames
        dt_df[1] = dt_df[1].apply(self._clean_str)

        # Put weights (even cols) and topics (odd cols) into separate dfs.
        weights_df = dt_df[dt_df.columns[2::2]]
        weights_df.index = dt_df[1]
        weights_df.columns = range(self.params["n_topics"])

        topics_df = dt_df[dt_df.columns[1:-1:2]]
        topics_df.index = dt_df[1]
        topics_df.columns = range(self.params["n_topics"])

        # Sort columns in weights_df separately for each row using topics_df.
        sorters_df = topics_df.apply(self._get_sort, axis=1)
        weights = weights_df.values
        sorters = np.vstack(sorters_df.values)
        # there has to be a better way to do this.
        for i in range(sorters.shape[0]):
            weights[i, :] = weights[i, sorters[i, :]]

        # Define topic names (e.g., topic_000)
        p_topic_g_doc_df = pd.DataFrame(columns=topic_names, data=weights, index=dt_df[1])
        p_topic_g_doc_df.index.name = "id"
        self.p_topic_g_doc_ = p_topic_g_doc_df.values
        self.p_topic_g_doc_df_ = p_topic_g_doc_df

        # Topic word weights
        p_word_g_topic_df = pd.read_csv(
            op.join(self.model_dir, "topic_word_weights.txt"),
            dtype=str,
            keep_default_na=False,
            na_values=[],
            sep="\t",
            header=None,
            names=["topic", "word", "weight"],
        )
        p_word_g_topic_df["weight"] = p_word_g_topic_df["weight"].astype(float)
        p_word_g_topic_df["topic"] = p_word_g_topic_df["topic"].astype(int)
        p_word_g_topic_df = p_word_g_topic_df.pivot(index="topic", columns="word", values="weight")
        p_word_g_topic_df = p_word_g_topic_df.div(p_word_g_topic_df.sum(axis=1), axis=0)
        self.p_word_g_topic_ = p_word_g_topic_df.values
        self.p_word_g_topic_df_ = p_word_g_topic_df

        # Remove all temporary files (text files, model, and outputs).
        shutil.rmtree(self.model_dir)

    def _clean_str(self, string):
        return op.basename(op.splitext(string)[0])

    def _get_sort(self, lst):
        return [i[0] for i in sorted(enumerate(lst), key=lambda x: x[1])]



"""Automated annotation of Cognitive Atlas labels."""
import logging
import re

import numpy as np
import pandas as pd

from .. import references
from ..due import due
from ..extract import download_cognitive_atlas
from ..utils import uk_to_us
from . import utils

LGR = logging.getLogger(__name__)


@due.dcite(references.COGNITIVE_ATLAS, description="Introduces the Cognitive Atlas.")
class CogAtLemmatizer(object):
    """Replace synonyms and abbreviations with Cognitive Atlas identifiers in text.

    Parameters
    ----------
    ontology_df : :obj:`pandas.DataFrame`, optional
        DataFrame with three columns (id, name, alias) and one row for each
        alias (e.g., synonym or abbreviation) for each term in the Cognitive
        Atlas. If None, loads ontology file from resources folder.

    Attributes
    ----------
    ontology_ : :obj:`pandas.DataFrame`
        Ontology in DataFrame form.
    regex_ : :obj:`dict`
        Dictionary linking aliases in ontology to regular expressions for
        lemmatization.

    References
    ----------
    * Poldrack, Russell A., et al. "The cognitive atlas: toward a
      knowledge foundation for cognitive neuroscience." Frontiers in
      neuroinformatics 5 (2011): 17. https://doi.org/10.3389/fninf.2011.00017
    """

    def __init__(self, ontology_df=None):
        if ontology_df is None:
            cogat = download_cognitive_atlas()
            self.ontology_ = pd.read_csv(cogat["ids"])
        else:
            assert isinstance(ontology_df, pd.DataFrame)
            self.ontology_ = ontology_df
        assert "id" in self.ontology_.columns
        assert "name" in self.ontology_.columns
        assert "alias" in self.ontology_.columns

        # Create regex dictionary
        regex_dict = {}
        for term in ontology_df["alias"].values:
            term_for_regex = term.replace("(", r"\(").replace(")", r"\)")
            regex = "\\b" + term_for_regex + "\\b"
            pattern = re.compile(regex, re.MULTILINE | re.IGNORECASE)
            regex_dict[term] = pattern
        self.regex_ = regex_dict

    def transform(self, text, convert_uk=True):
        """Replace terms in text with unique Cognitive Atlas identifiers.

        Parameters
        ----------
        text : :obj:`str`
            Text to convert.
        convert_uk : :obj:`bool`, optional
            Convert British English words in text to American English versions.
            Default is True.

        Returns
        -------
        text : :obj:`str`
            Text with Cognitive Atlas terms replaced with unique Cognitive
            Atlas identifiers.
        """
        if convert_uk:
            text = uk_to_us(text)

        for term_idx in self.ontology_.index:
            term = self.ontology_["alias"].loc[term_idx]
            term_id = self.ontology_["id"].loc[term_idx]
            text = re.sub(self.regex_[term], term_id, text)
        return text


@due.dcite(references.COGNITIVE_ATLAS, description="Introduces the Cognitive Atlas.")
def extract_cogat(text_df, id_df=None, text_column="abstract"):
    """Extract Cognitive Atlas terms and count instances using regular expressions.

    Parameters
    ----------
    text_df : (D x 2) :obj:`pandas.DataFrame`
        Pandas dataframe with at least two columns: 'id' and the text.
        D = document.
    id_df : (T x 3) :obj:`pandas.DataFrame`
        Cognitive Atlas ontology dataframe with at least three columns:
        'id' (unique identifier for term), 'alias' (natural language expression
        of term), and 'name' (preferred name of term; currently unused).
        T = term.
    text_column : :obj:`str`, optional
        Name of column in text_df that contains text. Default is 'abstract'.

    Returns
    -------
    counts_df : (D x T) :obj:`pandas.DataFrame`
        Term counts for documents in the corpus.
    rep_text_df : (D x 2) :obj:`pandas.DataFrame`
        Text DataFrame with terms replaced with their CogAt IDs.

    References
    ----------
    * Poldrack, Russell A., et al. "The cognitive atlas: toward a
      knowledge foundation for cognitive neuroscience." Frontiers in
      neuroinformatics 5 (2011): 17. https://doi.org/10.3389/fninf.2011.00017
    """
    text_df = text_df.copy()
    if id_df is None:
        cogat = download_cognitive_atlas()
        id_df = pd.read_csv(cogat["ids"])
    gazetteer = sorted(id_df["id"].unique().tolist())
    if "id" in text_df.columns:
        text_df.set_index("id", inplace=True)

    text_df[text_column] = text_df[text_column].fillna("")
    text_df[text_column] = text_df[text_column].apply(uk_to_us)

    # Create regex dictionary
    regex_dict = {}
    for term in id_df["alias"].values:
        term_for_regex = term.replace("(", r"\(").replace(")", r"\)")
        regex = "\\b" + term_for_regex + "\\b"
        pattern = re.compile(regex, re.MULTILINE | re.IGNORECASE)
        regex_dict[term] = pattern

    # Count
    count_arr = np.zeros((text_df.shape[0], len(gazetteer)), int)
    counts_df = pd.DataFrame(columns=gazetteer, index=text_df.index, data=count_arr)
    for term_idx in id_df.index:
        term = id_df["alias"].loc[term_idx]
        term_id = id_df["id"].loc[term_idx]
        pattern = regex_dict[term]
        counts_df[term_id] += text_df[text_column].str.count(pattern).astype(int)
        text_df[text_column] = text_df[text_column].str.replace(pattern, term_id)

    return counts_df, text_df


def expand_counts(counts_df, rel_df=None, weights=None):
    """Perform hierarchical expansion of counts across labels.

    Parameters
    ----------
    counts_df : (D x T) :obj:`pandas.DataFrame`
        Term counts for a corpus. T = term, D = document.
    rel_df : :obj:`pandas.DataFrame`
        Long-form DataFrame of term-term relationships with at least three columns:
        'input', 'output', and 'rel_type'.
    weights : :obj:`dict`
        Dictionary of weights per relationship type. E.g., {'isKind': 1}.
        Unspecified relationship types default to 0.

    Returns
    -------
    weighted_df : (D x T) :obj:`pandas.DataFrame`
        Term counts for a corpus after hierarchical expansion.
    """
    if rel_df is None:
        cogat = download_cognitive_atlas()
        rel_df = pd.read_csv(cogat["relationships"])
    weights_df = utils._generate_weights(rel_df, weights=weights)

    # First reorg counts_df so it has the same columns in the same order as
    # weight_df
    counts_columns = counts_df.columns.tolist()
    weights_columns = weights_df.columns.tolist()
    w_not_c = set(weights_columns) - set(counts_columns)
    c_not_w = set(counts_columns) - set(weights_columns)
    if c_not_w:
        raise Exception("Columns found in counts but not weights: {0}".format(", ".join(c_not_w)))

    for col in w_not_c:
        counts_df[col] = 0

    counts_df = counts_df[weights_columns]

    # Now matrix multiplication
    counts = counts_df.values
    weights = weights_df.values
    weighted = np.dot(counts, weights)
    weighted_df = pd.DataFrame(index=counts_df.index, columns=counts_df.columns, data=weighted)
    return weighted_df



"""Base classes for NiMARE."""
import gzip
import inspect
import logging
import multiprocessing as mp
import pickle
from abc import ABCMeta, abstractmethod
from collections import defaultdict
from hashlib import md5

import nibabel as nb
import numpy as np
from nilearn._utils.niimg_conversions import _check_same_fov
from nilearn.image import concat_imgs, resample_to_img

from .results import MetaResult
from .utils import get_masker

LGR = logging.getLogger(__name__)


class NiMAREBase(metaclass=ABCMeta):
    """Base class for NiMARE.

    TODO: Actually write/refactor class methods. They mostly come directly from sklearn
    https://github.com/scikit-learn/scikit-learn/blob/
    2a1e9686eeb203f5fddf44fd06414db8ab6a554a/sklearn/base.py#L141
    """

    def __init__(self):
        pass

    def _check_ncores(self, n_cores):
        """Check number of cores used for method."""
        if n_cores == -1:
            n_cores = mp.cpu_count()
        elif n_cores > mp.cpu_count():
            LGR.warning(
                "Desired number of cores ({0}) greater than number "
                "available ({1}). Setting to {1}.".format(n_cores, mp.cpu_count())
            )
            n_cores = mp.cpu_count()
        return n_cores

    @classmethod
    def _get_param_names(cls):
        """Get parameter names for the estimator."""
        # fetch the constructor or the original constructor before
        # deprecation wrapping if any
        init = getattr(cls.__init__, "deprecated_original", cls.__init__)
        if init is object.__init__:
            # No explicit constructor to introspect
            return []

        # introspect the constructor arguments to find the model parameters
        # to represent
        init_signature = inspect.signature(init)
        # Consider the constructor parameters excluding 'self'
        parameters = [
            p
            for p in init_signature.parameters.values()
            if p.name != "self" and p.kind != p.VAR_KEYWORD
        ]
        for p in parameters:
            if p.kind == p.VAR_POSITIONAL:
                raise RuntimeError(
                    "scikit-learn estimators should always "
                    "specify their parameters in the signature"
                    " of their __init__ (no varargs)."
                    " %s with constructor %s doesn't "
                    " follow this convention." % (cls, init_signature)
                )
        # Extract and sort argument names excluding 'self'
        return sorted([p.name for p in parameters])

    def get_params(self, deep=True):
        """Get parameters for this estimator.

        Parameters
        ----------
        deep : :obj:`bool`, optional
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.

        Returns
        -------
        params : :obj:`dict`
            Parameter names mapped to their values.
        """
        out = dict()
        for key in self._get_param_names():
            value = getattr(self, key, None)
            if deep and hasattr(value, "get_params"):
                deep_items = value.get_params().items()
                out.update((key + "__" + k, val) for k, val in deep_items)
            out[key] = value
        return out

    def set_params(self, **params):
        """Set the parameters of this estimator.

        The method works on simple estimators as well as on nested objects
        (such as pipelines). The latter have parameters of the form
        ``<component>__<parameter>`` so that it's possible to update each
        component of a nested object.

        Returns
        -------
        self
        """
        if not params:
            # Simple optimization to gain speed (inspect is slow)
            return self
        valid_params = self.get_params(deep=True)

        nested_params = defaultdict(dict)  # grouped by prefix
        for key, value in params.items():
            key, delim, sub_key = key.partition("__")
            if key not in valid_params:
                raise ValueError(
                    "Invalid parameter %s for estimator %s. "
                    "Check the list of available parameters "
                    "with `estimator.get_params().keys()`." % (key, self)
                )

            if delim:
                nested_params[key][sub_key] = value
            else:
                setattr(self, key, value)
                valid_params[key] = value

        for key, sub_params in nested_params.items():
            valid_params[key].set_params(**sub_params)

        return self

    def save(self, filename, compress=True):
        """Pickle the class instance to the provided file.

        Parameters
        ----------
        filename : :obj:`str`
            File to which object will be saved.
        compress : :obj:`bool`, optional
            If True, the file will be compressed with gzip. Otherwise, the
            uncompressed version will be saved. Default = True.
        """
        if compress:
            with gzip.GzipFile(filename, "wb") as file_object:
                pickle.dump(self, file_object)
        else:
            with open(filename, "wb") as file_object:
                pickle.dump(self, file_object)

    @classmethod
    def load(cls, filename, compressed=True):
        """Load a pickled class instance from file.

        Parameters
        ----------
        filename : :obj:`str`
            Name of file containing object.
        compressed : :obj:`bool`, optional
            If True, the file is assumed to be compressed and gzip will be used
            to load it. Otherwise, it will assume that the file is not
            compressed. Default = True.

        Returns
        -------
        obj : class object
            Loaded class object.
        """
        if compressed:
            try:
                with gzip.GzipFile(filename, "rb") as file_object:
                    obj = pickle.load(file_object)
            except UnicodeDecodeError:
                # Need to try this for python3
                with gzip.GzipFile(filename, "rb") as file_object:
                    obj = pickle.load(file_object, encoding="latin")
        else:
            try:
                with open(filename, "rb") as file_object:
                    obj = pickle.load(file_object)
            except UnicodeDecodeError:
                # Need to try this for python3
                with open(filename, "rb") as file_object:
                    obj = pickle.load(file_object, encoding="latin")

        if not isinstance(obj, cls):
            raise IOError("Pickled object must be {0}, not {1}".format(cls, type(obj)))

        return obj


class Estimator(NiMAREBase):
    """Estimators take in Datasets and return MetaResults."""

    # Inputs that must be available in input Dataset. Keys are names of
    # attributes to set; values are strings indicating location in Dataset.
    _required_inputs = {}

    def _validate_input(self, dataset):
        """Search for, and validate, required inputs as necessary."""
        if not hasattr(dataset, "slice"):
            raise ValueError(
                'Argument "dataset" must be a valid Dataset '
                "object, not a {0}".format(type(dataset))
            )

        if self._required_inputs:
            data = dataset.get(self._required_inputs)
            self.inputs_ = {}
            for k, v in data.items():
                if v is None:
                    raise ValueError(
                        "Estimator {0} requires input dataset to contain {1}, but "
                        "none were found.".format(self.__class__.__name__, k)
                    )
                self.inputs_[k] = v

    def _preprocess_input(self, dataset):
        """Perform any additional preprocessing steps on data in self.inputs_."""
        pass

    def fit(self, dataset):
        """Fit Estimator to Dataset.

        Parameters
        ----------
        dataset : :obj:`nimare.dataset.Dataset`
            Dataset object to analyze.

        Returns
        -------
        :obj:`nimare.results.MetaResult`
            Results of Estimator fitting.

        Notes
        -----
        The `fit` method is a light wrapper that runs input validation and
        preprocessing before fitting the actual model. Estimators' individual
        "fitting" methods are implemented as `_fit`, although users should
        call `fit`.
        """
        self._validate_input(dataset)
        self._preprocess_input(dataset)
        maps = self._fit(dataset)

        if hasattr(self, "masker") and self.masker is not None:
            masker = self.masker
        else:
            masker = dataset.masker

        self.results = MetaResult(self, masker, maps)
        return self.results

    @abstractmethod
    def _fit(self, dataset):
        """Apply estimation to dataset and output results.

        Must return a dictionary of results, where keys are names of images
        and values are ndarrays.
        """
        pass


class MetaEstimator(Estimator):
    """Base class for meta-analysis methods in :mod:`nimare.meta`."""

    def __init__(self, *args, **kwargs):
        mask = kwargs.get("mask")
        if mask is not None:
            mask = get_masker(mask)
        self.masker = mask

        self.resample = kwargs.get("resample", False)
        self.low_memory = kwargs.get("low_memory", False)

        # defaults for resampling images (nilearn's defaults do not work well)
        self._resample_kwargs = {"clip": True, "interpolation": "linear"}
        self._resample_kwargs.update(
            {k.split("resample__")[1]: v for k, v in kwargs.items() if k.startswith("resample__")}
        )

    def _preprocess_input(self, dataset):
        """Preprocess inputs to the Estimator from the Dataset as needed."""
        masker = self.masker or dataset.masker

        mask_img = masker.mask_img or masker.labels_img
        if isinstance(mask_img, str):
            mask_img = nb.load(mask_img)

        for name, (type_, _) in self._required_inputs.items():
            if type_ == "image":
                # If no resampling is requested, check if resampling is required
                if not self.resample:
                    check_imgs = {img: nb.load(img) for img in self.inputs_[name]}
                    _check_same_fov(**check_imgs, reference_masker=mask_img, raise_error=True)
                    imgs = list(check_imgs.values())
                else:
                    # resampling will only occur if shape/affines are different
                    # making this harmless if all img shapes/affines are the same
                    # as the reference
                    imgs = [
                        resample_to_img(nb.load(img), mask_img, **self._resample_kwargs)
                        for img in self.inputs_[name]
                    ]

                # input to NiFtiLabelsMasker must be 4d
                img4d = concat_imgs(imgs, ensure_ndim=4)

                # Mask required input images using either the dataset's mask or
                # the estimator's.
                temp_arr = masker.transform(img4d)

                # An intermediate step to mask out bad voxels. Can be dropped
                # once PyMARE is able to handle masked arrays or missing data.
                bad_voxel_idx = np.where(temp_arr == 0)[1]
                bad_voxel_idx = np.unique(bad_voxel_idx)
                LGR.debug('Masking out {} "bad" voxels'.format(len(bad_voxel_idx)))
                temp_arr[:, bad_voxel_idx] = 0

                self.inputs_[name] = temp_arr
            elif type_ == "coordinates":
                # Try to load existing MA maps
                if hasattr(self, "kernel_transformer"):
                    self.kernel_transformer._infer_names(affine=md5(mask_img.affine).hexdigest())
                    if self.kernel_transformer.image_type in dataset.images.columns:
                        files = dataset.get_images(
                            ids=dataset.ids,
                            imtype=self.kernel_transformer.image_type,
                        )
                        if all(f is not None for f in files):
                            self.inputs_["ma_maps"] = files

                # Set the coordinates directly as well
                self.inputs_[name] = dataset.coordinates.copy()


class Transformer(NiMAREBase):
    """Transformers take in Datasets and return Datasets.

    Initialize with hyperparameters.
    """

    def __init__(self):
        pass

    @abstractmethod
    def transform(self, dataset):
        """Add stuff to transformer."""
        if not hasattr(dataset, "slice"):
            raise ValueError(
                'Argument "dataset" must be a valid Dataset '
                "object, not a {0}".format(type(dataset))
            )


class Decoder(NiMAREBase):
    """Base class for decoders in :mod:`nimare.decode`."""

    __id_cols = ["id", "study_id", "contrast_id"]

    def _preprocess_input(self, dataset):
        """Select features for model based on requested features and feature_group.

        This also takes into account which features have at least one study in the
        Dataset with the feature.
        """
        # Reduce feature list as desired
        if self.feature_group is not None:
            if not self.feature_group.endswith("__"):
                self.feature_group += "__"
            feature_names = dataset.annotations.columns.values
            feature_names = [f for f in feature_names if f.startswith(self.feature_group)]
            if self.features is not None:
                features = [f.split("__")[-1] for f in feature_names if f in self.features]
            else:
                features = feature_names
        else:
            if self.features is None:
                features = dataset.annotations.columns.values
            else:
                features = self.features
        features = [f for f in features if f not in self.__id_cols]
        # At least one study in the dataset much have each label
        counts = (dataset.annotations[features] > self.frequency_threshold).sum(0)
        features = counts[counts > 0].index.tolist()
        if not len(features):
            raise Exception("No features identified in Dataset!")
        self.features_ = features

    def fit(self, dataset):
        """Fit Decoder to Dataset.

        Parameters
        ----------
        dataset : :obj:`nimare.dataset.Dataset`
            Dataset object to analyze.

        Returns
        -------
        :obj:`nimare.results.MetaResult`
            Results of Decoder fitting.

        Notes
        -----
        The `fit` method is a light wrapper that runs input validation and
        preprocessing before fitting the actual model. Decoders' individual
        "fitting" methods are implemented as `_fit`, although users should
        call `fit`.

        Selection of features based on requested features and feature group is performed in
        `Decoder._preprocess_input`.
        """
        self._preprocess_input(dataset)
        self._fit(dataset)

    @abstractmethod
    def _fit(self, dataset):
        """Apply decoding to dataset and output results.

        Must return a DataFrame, with one row for each feature.
        """
        pass



"""Test nimare.utils."""
import logging
import os
import os.path as op

import nibabel as nib
import numpy as np
import pytest

from nimare import utils


def test_find_stem():
    """Test nimare.utils.find_stem."""
    test_array = [
        "/home/data/dataset/file1.nii.gz",
        "/home/data/dataset/file2.nii.gz",
        "/home/data/dataset/file3.nii.gz",
        "/home/data/dataset/file4.nii.gz",
        "/home/data/dataset/file5.nii.gz",
    ]
    stem = utils.find_stem(test_array)
    assert stem == "/home/data/dataset/file"

    test_array = [
        "/home/data/dataset/subfolder1/file1.nii.gz",
        "/home/data/dataset/subfolder1/file2.nii.gz",
        "/home/data/dataset/subfolder2/file3.nii.gz",
        "/home/data/dataset/subfolder2/file4.nii.gz",
        "/home/data/dataset/subfolder3/file5.nii.gz",
    ]
    stem = utils.find_stem(test_array)
    assert stem == "/home/data/dataset/subfolder"

    test_array = [
        "/home/data/file1_test-filename_test.nii.gz",
        "/home/data/file2_test-filename_test.nii.gz",
        "/home/data/file3_test-filename_test.nii.gz",
        "/home/data/file4_test-filename_test.nii.gz",
        "/home/data/file5_test-filename_test.nii.gz",
    ]
    stem = utils.find_stem(test_array)
    assert stem == "/home/data/file"

    test_array = [
        "souse",
        "youse",
        "house",
        "mouse",
        "louse",
    ]
    stem = utils.find_stem(test_array)
    assert stem == ""


def test_get_template():
    """Test nimare.utils.get_template."""
    img = utils.get_template(space="mni152_1mm", mask=None)
    assert isinstance(img, nib.Nifti1Image)
    img = utils.get_template(space="mni152_1mm", mask="brain")
    assert isinstance(img, nib.Nifti1Image)
    img = utils.get_template(space="mni152_1mm", mask="gm")
    assert isinstance(img, nib.Nifti1Image)
    img = utils.get_template(space="mni152_2mm", mask=None)
    assert isinstance(img, nib.Nifti1Image)
    img = utils.get_template(space="mni152_2mm", mask="brain")
    assert isinstance(img, nib.Nifti1Image)
    img = utils.get_template(space="mni152_2mm", mask="gm")
    assert isinstance(img, nib.Nifti1Image)


def test_get_resource_path():
    """Test nimare.utils.get_resource_path."""
    print(utils.get_resource_path())
    assert op.isdir(utils.get_resource_path())


@pytest.mark.parametrize(
    "has_low_memory,low_memory",
    [
        (True, True),
        (False, False),
    ],
)
def test_use_memmap(caplog, has_low_memory, low_memory):
    """Test the memmapping decorator."""
    LGR = logging.getLogger(__name__)

    class DummyClass:
        def __init__(self, has_low_memory, low_memory):
            self.has_low_memory = has_low_memory
            if has_low_memory:
                self.low_memory = low_memory

        @utils.use_memmap(LGR)
        def test_decorator(self):
            assert hasattr(self, "memmap_filenames")
            if self.has_low_memory:
                assert hasattr(self, "low_memory")
                if self.low_memory:
                    assert os.path.isfile(self.memmap_filenames[0])
                else:
                    assert self.memmap_filenames[0] is None
            return self.memmap_filenames

        @utils.use_memmap(LGR)
        def bad_justin_timberlake(self):
            raise ValueError("It's gonna be may!")

    my_class = DummyClass(has_low_memory, low_memory)

    # make sure memmap file has been deleted
    my_class.test_decorator()
    first_memmap_filename = my_class.memmap_filenames[0]

    # run bad function
    with pytest.raises(ValueError):
        my_class.bad_justin_timberlake()
    assert "failed, removing" in caplog.text

    if hasattr(my_class, "low_memory") and my_class.low_memory:
        assert not os.path.isfile(first_memmap_filename)
        assert not os.path.isfile(my_class.memmap_filenames[0])
        # test when a function is called a new memmap file is created
        assert first_memmap_filename != my_class.memmap_filenames[0]


def test_tal2mni():
    """TODO: Get converted coords from official site."""
    test = np.array([[-44, 31, 27], [20, -32, 14], [28, -76, 28]])
    true = np.array(
        [
            [-45.83997568, 35.97904559, 23.55194326],
            [22.69248975, -31.34145016, 13.91284087],
            [31.53113226, -76.61685748, 33.22105166],
        ]
    )
    assert np.allclose(utils.tal2mni(test), true)


def test_mni2tal():
    """TODO: Get converted coords from official site."""
    test = np.array([[-44, 31, 27], [20, -32, 14], [28, -76, 28]])
    true = np.array(
        [[-42.3176, 26.0594, 29.7364], [17.4781, -32.6076, 14.0009], [24.7353, -75.0184, 23.3283]]
    )
    assert np.allclose(utils.mni2tal(test), true)


def test_vox2mm():
    """Test vox2mm."""
    test = np.array([[20, 20, 20], [0, 0, 0]])
    true = np.array([[50.0, -86.0, -32.0], [90.0, -126.0, -72.0]])
    img = utils.get_template(space="mni152_2mm", mask=None)
    aff = img.affine
    assert np.array_equal(utils.vox2mm(test, aff), true)


def test_mm2vox():
    """Test mm2vox."""
    test = np.array([[20, 20, 20], [0, 0, 0]])
    true = np.array([[35.0, 73.0, 46.0], [45.0, 63.0, 36.0]])
    img = utils.get_template(space="mni152_2mm", mask=None)
    aff = img.affine
    assert np.array_equal(utils.mm2vox(test, aff), true)



"""Test estimator, kerneltransformer, and multiple comparisons corrector performance."""
import os
from contextlib import ExitStack as does_not_raise

import nibabel as nib
import numpy as np
import pytest

from ..correct import FDRCorrector, FWECorrector
from ..generate import create_coordinate_dataset
from ..meta import ale, kernel, mkda
from ..meta.utils import compute_kda_ma
from ..results import MetaResult
from ..utils import mm2vox

# set significance levels used for testing.
ALPHA = 0.05

if os.environ.get("CIRCLECI"):
    N_CORES = 1
else:
    N_CORES = -1


# PRECOMPUTED FIXTURES
# --------------------

##########################################
# random state
##########################################
@pytest.fixture(scope="session")
def random():
    """Set random state for the tests."""
    np.random.seed(1939)


##########################################
# simulated dataset(s)
##########################################
@pytest.fixture(
    scope="session",
    params=[
        pytest.param(
            {
                "foci": 5,
                "fwhm": 10.0,
                "n_studies": 40,
                "sample_size": 30,
                "n_noise_foci": 20,
                "seed": 1939,
            },
            id="normal_data",
        )
    ],
)
def simulatedata_cbma(request):
    """Set the simulated CBMA data according to parameters."""
    return request.param["fwhm"], create_coordinate_dataset(**request.param)


##########################################
# signal and non-signal masks
##########################################
@pytest.fixture(scope="session")
def signal_masks(simulatedata_cbma):
    """Define masks of signal and non-signal for performance evaluation."""
    _, (ground_truth_foci, dataset) = simulatedata_cbma
    ground_truth_foci_ijks = [
        tuple(mm2vox(focus, dataset.masker.mask_img.affine)) for focus in ground_truth_foci
    ]
    return _create_signal_mask(np.array(ground_truth_foci_ijks), dataset.masker.mask_img)


##########################################
# meta-analysis estimators
##########################################
@pytest.fixture(
    scope="session",
    params=[
        pytest.param(ale.ALE, id="ale"),
        pytest.param(mkda.MKDADensity, id="mkda"),
        pytest.param(mkda.KDA, id="kda"),
    ],
)
def meta_est(request):
    """Define meta-analysis estimators for tests."""
    return request.param


##########################################
# meta-analysis estimator parameters
##########################################
@pytest.fixture(
    scope="session",
    params=[
        pytest.param({"null_method": "montecarlo", "n_iters": 100}, id="montecarlo"),
        pytest.param({"null_method": "approximate"}, id="approximate"),
        pytest.param(
            {"null_method": "reduced_montecarlo", "n_iters": 100}, id="reduced_montecarlo"
        ),
    ],
)
def meta_params(request):
    """Define meta-analysis Estimator parameters for tests."""
    return request.param


##########################################
# meta-analysis kernels
##########################################
@pytest.fixture(
    scope="session",
    params=[
        pytest.param(kernel.ALEKernel, id="ale_kernel"),
        pytest.param(kernel.MKDAKernel, id="mkda_kernel"),
        pytest.param(kernel.KDAKernel, id="kda_kernel"),
    ],
)
def kern(request):
    """Define kernel transformers for tests."""
    return request.param


##########################################
# multiple comparison correctors (testing)
##########################################
@pytest.fixture(
    scope="session",
    params=[
        pytest.param(FWECorrector(method="bonferroni"), id="fwe_bonferroni"),
        pytest.param(
            FWECorrector(method="montecarlo", voxel_thresh=ALPHA, n_iters=100, n_cores=N_CORES),
            id="fwe_montecarlo",
        ),
        pytest.param(FDRCorrector(method="indep", alpha=ALPHA), id="fdr_indep"),
        pytest.param(FDRCorrector(method="negcorr", alpha=ALPHA), id="fdr_negcorr"),
    ],
)
def corr(request):
    """Define multiple comparisons correctors for tests."""
    return request.param


##########################################
# multiple comparison correctors (smoke)
##########################################
@pytest.fixture(
    scope="session",
    params=[
        pytest.param(FWECorrector(method="bonferroni"), id="fwe_bonferroni"),
        pytest.param(
            FWECorrector(method="montecarlo", voxel_thresh=ALPHA, n_iters=2, n_cores=1),
            id="fwe_montecarlo",
        ),
        pytest.param(FDRCorrector(method="indep", alpha=ALPHA), id="fdr_indep"),
        pytest.param(FDRCorrector(method="negcorr", alpha=ALPHA), id="fdr_negcorr"),
    ],
)
def corr_small(request):
    """Define multiple comparisons correctors for tests."""
    return request.param


###########################################
# all meta-analysis estimator/kernel combos
###########################################
@pytest.fixture(scope="session")
def meta(simulatedata_cbma, meta_est, kern, meta_params):
    """Define estimator/kernel combinations for tests."""
    fwhm, (_, _) = simulatedata_cbma
    if kern == kernel.KDAKernel or kern == kernel.MKDAKernel:
        kern = kern(r=fwhm / 2)
    else:
        kern = kern()

    # instantiate meta-analysis estimator
    return meta_est(kern, **meta_params)


###########################################
# meta-analysis estimator results
###########################################
@pytest.fixture(scope="session")
def meta_res(simulatedata_cbma, meta, random):
    """Define estimators for tests."""
    _, (_, dataset) = simulatedata_cbma
    # CHECK IF META/KERNEL WORK TOGETHER
    ####################################
    meta_expectation = does_not_raise()

    with meta_expectation:
        res = meta.fit(dataset)
    # if creating the result failed (expected), do not continue
    if isinstance(meta_expectation, type(pytest.raises(ValueError))):
        pytest.xfail("this meta-analysis & kernel combo fails")
    # instantiate meta-analysis estimator
    return res


###########################################
# corrected results (testing)
###########################################
@pytest.fixture(scope="session")
def meta_cres(meta, meta_res, corr, random):
    """Define corrected results for tests."""
    return _transform_res(meta, meta_res, corr)


###########################################
# corrected results (smoke)
###########################################
@pytest.fixture(scope="session")
def meta_cres_small(meta, meta_res, corr_small, random):
    """Define corrected results for tests."""
    return _transform_res(meta, meta_res, corr_small)


# --------------
# TEST FUNCTIONS
# --------------


def test_meta_fit_smoke(meta_res):
    """Smoke test for meta-analytic estimator fit."""
    assert isinstance(meta_res, MetaResult)


@pytest.mark.performance
def test_meta_fit_performance(meta_res, signal_masks, simulatedata_cbma):
    """Test meta-analytic estimator fit performance."""
    _, (ground_truth_foci, _) = simulatedata_cbma
    mask = meta_res.masker.mask_img
    ground_truth_foci_ijks = [tuple(mm2vox(focus, mask.affine)) for focus in ground_truth_foci]
    sig_idx, nonsig_idx = [
        meta_res.masker.transform(img).astype(bool).squeeze() for img in signal_masks
    ]

    # all estimators generate p-values
    p_array = meta_res.get_map("p", return_type="array")

    # poor performer(s)
    if (
        isinstance(meta_res.estimator, ale.ALE)
        and isinstance(meta_res.estimator.kernel_transformer, kernel.KDAKernel)
        and meta_res.estimator.get_params().get("null_method") == "approximate"
    ):
        good_sensitivity = True
        good_specificity = False
    elif (
        isinstance(meta_res.estimator, ale.ALE)
        and isinstance(meta_res.estimator.kernel_transformer, kernel.KDAKernel)
        and "montecarlo" in meta_res.estimator.get_params().get("null_method")
    ):
        good_sensitivity = False
        good_specificity = True
    elif (
        isinstance(meta_res.estimator, ale.ALE)
        and type(meta_res.estimator.kernel_transformer) == kernel.KDAKernel
        and "montecarlo" in meta_res.estimator.get_params().get("null_method")
    ):
        good_sensitivity = False
        good_specificity = True
    elif (
        isinstance(meta_res.estimator, ale.ALE)
        and type(meta_res.estimator.kernel_transformer) == kernel.KDAKernel
        and meta_res.estimator.get_params().get("null_method") == "approximate"
    ):
        good_sensitivity = True
        good_specificity = False
    elif (
        isinstance(meta_res.estimator, mkda.MKDADensity)
        and isinstance(meta_res.estimator.kernel_transformer, kernel.ALEKernel)
        and meta_res.estimator.get_params().get("null_method") != "reduced_montecarlo"
    ):
        good_sensitivity = False
        good_specificity = True
    else:
        good_sensitivity = True
        good_specificity = True

    _check_p_values(
        p_array,
        meta_res.masker,
        sig_idx,
        nonsig_idx,
        ALPHA,
        ground_truth_foci_ijks,
        n_iters=None,
        good_sensitivity=good_sensitivity,
        good_specificity=good_specificity,
    )


def test_corr_transform_smoke(meta_cres_small):
    """Smoke test for corrector transform."""
    assert isinstance(meta_cres_small, MetaResult)


@pytest.mark.performance
def test_corr_transform_performance(meta_cres, corr, signal_masks, simulatedata_cbma):
    """Test corrector transform performance."""
    _, (ground_truth_foci, _) = simulatedata_cbma
    mask = meta_cres.masker.mask_img
    ground_truth_foci_ijks = [tuple(mm2vox(focus, mask.affine)) for focus in ground_truth_foci]
    sig_idx, nonsig_idx = [
        meta_cres.masker.transform(img).astype(bool).squeeze() for img in signal_masks
    ]

    p_array = meta_cres.maps.get("p")
    if p_array is None or corr.method == "montecarlo":
        p_array = 10 ** -meta_cres.maps.get("logp_level-voxel_corr-FWE_method-montecarlo")

    n_iters = corr.parameters.get("n_iters")

    # ALE with MKDA kernel with montecarlo correction
    # combination gives poor performance
    if (
        isinstance(meta_cres.estimator, ale.ALE)
        and isinstance(meta_cres.estimator.kernel_transformer, kernel.MKDAKernel)
        and meta_cres.estimator.get_params().get("null_method") == "approximate"
        and corr.method != "montecarlo"
    ):
        good_sensitivity = True
        good_specificity = False
    elif (
        isinstance(meta_cres.estimator, ale.ALE)
        and isinstance(meta_cres.estimator.kernel_transformer, kernel.MKDAKernel)
        and "montecarlo" in meta_cres.estimator.get_params().get("null_method")
    ):
        good_sensitivity = False
        good_specificity = True
    elif (
        isinstance(meta_cres.estimator, ale.ALE)
        and isinstance(meta_cres.estimator.kernel_transformer, kernel.MKDAKernel)
        and meta_cres.estimator.get_params().get("null_method") == "approximate"
        and corr.method == "montecarlo"
    ):
        good_sensitivity = False
        good_specificity = True
    elif (
        isinstance(meta_cres.estimator, ale.ALE)
        and type(meta_cres.estimator.kernel_transformer) == kernel.KDAKernel
        and (
            "montecarlo" in meta_cres.estimator.get_params().get("null_method")
            or (
                meta_cres.estimator.get_params().get("null_method") == "approximate"
                and corr.method == "montecarlo"
            )
        )
    ):
        good_sensitivity = False
        good_specificity = True
    elif (
        isinstance(meta_cres.estimator, ale.ALE)
        and type(meta_cres.estimator.kernel_transformer) == kernel.KDAKernel
        and meta_cres.estimator.get_params().get("null_method") == "approximate"
    ):
        good_sensitivity = True
        good_specificity = False
    elif (
        isinstance(meta_cres.estimator, mkda.MKDADensity)
        and isinstance(meta_cres.estimator.kernel_transformer, kernel.ALEKernel)
        and meta_cres.estimator.get_params().get("null_method") != "reduced_montecarlo"
        and corr.method != "montecarlo"
    ):
        good_sensitivity = False
        good_specificity = True
    else:
        good_sensitivity = True
        good_specificity = True

    _check_p_values(
        p_array,
        meta_cres.masker,
        sig_idx,
        nonsig_idx,
        ALPHA,
        ground_truth_foci_ijks,
        n_iters=n_iters,
        good_sensitivity=good_sensitivity,
        good_specificity=good_specificity,
    )


# -----------------
# UTILITY FUNCTIONS
# -----------------


def _create_signal_mask(ground_truth_foci_ijks, mask):
    """Create complementary binary images to identify areas of likely significance and nonsignificance.

    Parameters
    ----------
    ground_truth_foci_ijks : array_like
        Ground truth ijk coordinates of foci.
    mask : :obj:`nibabel.Nifti1Image`
        Input mask to define shape and size of output binary masks

    Returns
    -------
    sig_map : :obj:`nibabel.Nifti1Image`
        Binary image representing regions around the
        ground truth foci expected to be significant.
    nonsig_map : :obj:`nibabel.Nifti1Image`
        Binary image representing regions not expected
        to be significant within the brain.
    """
    dims = mask.shape
    vox_dims = mask.header.get_zooms()

    # area where I'm reasonably certain there are significant results
    sig_prob_map = compute_kda_ma(
        dims, vox_dims, ground_truth_foci_ijks, r=2, value=1, sum_overlap=False
    )

    # area where I'm reasonably certain there are not significant results
    nonsig_prob_map = compute_kda_ma(
        dims, vox_dims, ground_truth_foci_ijks, r=14, value=1, sum_overlap=False
    )
    sig_map = nib.Nifti1Image((sig_prob_map == 1).astype(int), affine=mask.affine)
    nonsig_map = nib.Nifti1Image((nonsig_prob_map == 0).astype(int), affine=mask.affine)
    return sig_map, nonsig_map


def _check_p_values(
    p_array,
    masker,
    sig_idx,
    nonsig_idx,
    alpha,
    ground_truth_foci_ijks,
    n_iters=None,
    good_sensitivity=True,
    good_specificity=True,
):
    """Check if p-values are within the correct range."""
    ################################################
    # CHECK IF P-VALUES ARE WITHIN THE CORRECT RANGE
    ################################################
    if n_iters:
        assert p_array.min() >= (1.0 / n_iters)
        assert p_array.max() <= 1.0 - (1.0 / n_iters)
    else:
        assert (p_array >= 0).all() and (p_array <= 1).all()

    p_map = masker.inverse_transform(p_array).get_fdata()

    # reformat coordinate indices to index p_map
    gtf_idx = [
        [ground_truth_foci_ijks[i][j] for i in range(len(ground_truth_foci_ijks))]
        for j in range(3)
    ]

    best_chance_p_values = p_map[gtf_idx]
    assert all(best_chance_p_values < ALPHA) == good_sensitivity

    p_array_sig = p_array[sig_idx]
    p_array_nonsig = p_array[nonsig_idx]

    # assert that at least 50% of voxels surrounding the foci
    # are significant at alpha = .05
    observed_sig = p_array_sig < alpha
    observed_sig_perc = observed_sig.sum() / len(observed_sig)
    assert (observed_sig_perc >= 0.5) == good_sensitivity

    # assert that more than 95% of voxels farther away
    # from foci are nonsignificant at alpha = 0.05
    observed_nonsig = p_array_nonsig > alpha
    observed_nonsig_perc = observed_nonsig.sum() / len(observed_nonsig)
    assert np.isclose(observed_nonsig_perc, (1 - alpha), atol=0.05) == good_specificity


def _transform_res(meta, meta_res, corr):
    """Evaluate whether meta estimator and corrector work together."""
    #######################################
    # CHECK IF META/CORRECTOR WORK TOGETHER
    #######################################
    # all combinations of meta-analysis estimators and multiple comparison correctors
    # that do not work together
    corr_expectation = does_not_raise()

    with corr_expectation:
        cres = corr.transform(meta_res)

    # if multiple correction failed (expected) do not continue
    if isinstance(corr_expectation, type(pytest.raises(ValueError))):
        pytest.xfail("this meta-analysis & corrector combo fails")
    return cres



"""Run a contrast permutation meta-analysis on a set of images."""
import logging
import os
import pathlib

import numpy as np
from nilearn.masking import apply_mask
from nilearn.mass_univariate import permuted_ols

from ..results import MetaResult
from ..utils import get_template

LGR = logging.getLogger(__name__)


def conperm_workflow(contrast_images, mask_image=None, output_dir=None, prefix="", n_iters=10000):
    """Run a contrast permutation workflow."""
    if mask_image is None:
        target = "mni152_2mm"
        mask_image = get_template(target, mask="brain")

    n_studies = len(contrast_images)
    LGR.info("Loading contrast maps...")
    z_data = apply_mask(contrast_images, mask_image)

    boilerplate = """
A contrast permutation analysis was performed on a sample of {n_studies}
images. A brain mask derived from the MNI 152 template (Fonov et al., 2009;
Fonov et al., 2011) was applied at 2x2x2mm resolution. The sign flipping
method used was implemented as described in Maumet & Nichols (2016), with
{n_iters} iterations used to estimate the null distribution.

References
----------
- Fonov, V., Evans, A. C., Botteron, K., Almli, C. R., McKinstry, R. C.,
Collins, D. L., & Brain Development Cooperative Group. (2011).
Unbiased average age-appropriate atlases for pediatric studies.
Neuroimage, 54(1), 313-327.
- Fonov, V. S., Evans, A. C., McKinstry, R. C., Almli, C. R., & Collins, D. L.
(2009). Unbiased nonlinear average age-appropriate brain templates from birth
to adulthood. NeuroImage, (47), S102.
- Maumet, C., & Nichols, T. E. (2016). Minimal Data Needed for Valid & Accurate
Image-Based fMRI Meta-Analysis. https://doi.org/10.1101/048249
    """

    LGR.info("Performing meta-analysis.")
    log_p_map, t_map, _ = permuted_ols(
        np.ones((z_data.shape[0], 1)),
        z_data,
        confounding_vars=None,
        model_intercept=False,  # modeled by tested_vars
        n_perm=n_iters,
        two_sided_test=True,
        random_state=42,
        n_jobs=1,
        verbose=0,
    )
    res = {"logp": log_p_map, "t": t_map}
    # The t_test function will stand in for the Estimator in the results object
    res = MetaResult(permuted_ols, mask_image, maps=res)

    boilerplate = boilerplate.format(n_studies=n_studies, n_iters=n_iters)

    if output_dir is None:
        output_dir = os.getcwd()
    else:
        pathlib.Path(output_dir).mkdir(parents=True, exist_ok=True)

    LGR.info("Saving output maps...")
    res.save_maps(output_dir=output_dir, prefix=prefix)
    LGR.info("Workflow completed.")
    LGR.info(boilerplate)



"""Tools for managing meta-analytic results."""
import copy
import logging
import os

from nibabel.funcs import squeeze_image

from .utils import get_masker

LGR = logging.getLogger(__name__)


class MetaResult(object):
    """Base class for meta-analytic results.

    Parameters
    ----------
    estimator : :class:`nimare.base.Estimator`
        The Estimator used to generate the maps in the MetaResult.
    mask : Niimg-like or `nilearn.input_data.base_masker.BaseMasker`
        Mask for converting maps between arrays and images.
    maps : :obj:`dict` or None, optional
        Maps to store in the object. Default is None.

    Attributes
    ----------
    estimator : :class:`nimare.base.Estimator`
        The Estimator used to generate the maps in the MetaResult.
    masker : :class:`nilearn.input_data.NiftiMasker` or similar
        Masker object.
    maps : :obj:`dict`
        Keys are map names and values are arrays.
    """

    def __init__(self, estimator, mask, maps=None):
        self.estimator = estimator
        self.masker = get_masker(mask)
        self.maps = maps or {}

    def get_map(self, name, return_type="image"):
        """Get stored map as image or array.

        Parameters
        ----------
        name : :obj:`str`
            Name of the map. Used to index self.maps.
        return_type : {'image', 'array'}, optional
            Whether to return a niimg ('image') or a numpy array.
            Default is 'image'.
        """
        m = self.maps.get(name)
        if m is None:
            raise ValueError("No map with name '{}' found.".format(name))
        if return_type == "image":
            # pending resolution of https://github.com/nilearn/nilearn/issues/2724
            try:
                return self.masker.inverse_transform(m)
            except IndexError:
                return squeeze_image(self.masker.inverse_transform([m]))
        return m

    def save_maps(self, output_dir=".", prefix="", prefix_sep="_", names=None):
        """Save results to files.

        Parameters
        ----------
        output_dir : :obj:`str`, optional
            Output directory in which to save results. If the directory doesn't
            exist, it will be created. Default is current directory.
        prefix : :obj:`str`, optional
            Prefix to prepend to output file names.
            Default is None.
        prefix_sep : :obj:`str`, optional
            Separator to add between prefix and default file names.
            Default is _.
        names : None or :obj:`list` of :obj:`str`, optional
            Names of specific maps to write out. If None, save all maps.
            Default is None.
        """
        if prefix == "":
            prefix_sep = ""

        if not prefix.endswith(prefix_sep):
            prefix = prefix + prefix_sep

        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        names = names or list(self.maps.keys())
        maps = {k: self.get_map(k) for k in names}

        for imgtype, img in maps.items():
            filename = prefix + imgtype + ".nii.gz"
            outpath = os.path.join(output_dir, filename)
            img.to_filename(outpath)

    def copy(self):
        """Return copy of result object."""
        new = MetaResult(self.estimator, self.masker, copy.deepcopy(self.maps))
        return new



"""References to be imported and injected at relevant places throughout the library."""
from .due import BibTeX, Doi

TEXT2BRAIN = Doi("https://doi.org/10.1007/978-3-030-00931-1_67")

WORD2BRAIN = Doi("10.1101/299024")

BOLTZMANNMODEL = BibTeX(
    """
    @article{DBLP:journals/corr/MontiLLAM16,
      author    = {Ricardo Pio Monti and Romy Lorenz and Robert Leech and
                   Christoforos Anagnostopoulos and Giovanni Montana},
      title     = {Text-mining the NeuroSynth corpus using Deep Boltzmann
                   Machines},
      journal   = {CoRR},
      volume    = {abs/1605.00223},
      year      = {2016},
      url       = {http://arxiv.org/abs/1605.00223},
      archivePrefix = {arXiv},
      eprint    = {1605.00223},
      timestamp = {Wed, 07 Jun 2017 14:42:40 +0200},
      biburl    = {https://dblp.org/rec/bib/journals/corr/MontiLLAM16},
      bibsource = {dblp computer science bibliography, https://dblp.org}}
    """
)

GCLDAMODEL = Doi("10.1371/journal.pcbi.1005649")

LDA = BibTeX(
    """
    @article{blei2003latent,
      title={Latent dirichlet allocation},
      author={Blei, David M and Ng, Andrew Y and Jordan, Michael I},
      journal={Journal of machine Learning research},
      volume={3},
      number={Jan},
      pages={993--1022},
      year={2003}}
    """
)

MALLET = BibTeX(
    """
    @article{mallettoolbox,
      title={MALLET: A Machine Learning for Language Toolkit.},
      author={McCallum, Andrew K},
      year={2002}}
    """
)

LDAMODEL = Doi("10.1371/journal.pcbi.1002707")

COGNITIVE_ATLAS = Doi("10.3389/fninf.2011.00017")

COGNITIVE_PARADIGM_ONTOLOGY = Doi("10.1007/s12021-011-9126-x")

ATHENA = Doi("10.3389/fnins.2019.00494")

LOG_LIKELIHOOD = Doi("10.1145/1577069.1755845")

GCLDA_DECODING = Doi("10.1371/journal.pcbi.1005649")

NEUROSYNTH = Doi("10.1038/nmeth.1635")

BRAINMAP_DECODING = Doi("10.1007/s00429-013-0698-0")

ALE1 = BibTeX(
    """
    @article{turkeltaub2002meta,
      title={Meta-analysis of the functional neuroanatomy of single-word
             reading: method and validation},
      author={Turkeltaub, Peter E and Eden, Guinevere F and Jones,
              Karen M and Zeffiro, Thomas A},
      journal={Neuroimage},
      volume={16},
      number={3},
      pages={765--780},
      year={2002},
      publisher={Elsevier}
    }
    """
)

ALE2 = Doi("10.1002/hbm.21186")

ALE3 = Doi("10.1016/j.neuroimage.2011.09.017")

ALE_KERNEL = Doi("10.1002/hbm.20718")

SCALE = Doi("10.1016/j.neuroimage.2014.06.007")

MKDA = Doi("10.1093/scan/nsm015")

KDA1 = Doi("10.1016/S1053-8119(03)00078-8")

KDA2 = Doi("10.1016/j.neuroimage.2004.03.052")

BHICP = Doi("10.1198/jasa.2011.ap09735")

HPGRF = BibTeX(
    """
    @article{kang2014bayesian,
      title={A Bayesian hierarchical spatial point process model for
             multi-type neuroimaging meta-analysis},
      author={Kang, Jian and Nichols, Thomas E and Wager, Tor D and
              Johnson, Timothy D},
      journal={The annals of applied statistics},
      volume={8},
      number={3},
      pages={1800},
      year={2014},
      publisher={NIH Public Access}
      }
    """
)

SBLFR = Doi("10.1111/biom.12713")

SBR = Doi("10.1214/11-AOAS523")

PEAKS2MAPS = Doi("10.7490/f1000research.1116395.1")

FISHERS = BibTeX(
    """
    @article{fisher1932statistical,
       title={Statistical methods for research workers, Edinburgh:
              Oliver and Boyd, 1925},
       author={Fisher, RA},
       journal={Google Scholar},
       year={1932}
       }
    """
)

STOUFFERS = BibTeX(
    """
    @article{stouffer1949american,
      title={The American soldier: Adjustment during army life. (Studies
             in social psychology in World War II), Vol. 1},
      author={Stouffer, Samuel A and Suchman, Edward A and DeVinney,
              Leland C and Star, Shirley A and Williams Jr, Robin M},
      year={1949},
      publisher={Princeton Univ. Press}
      }
    """
)

WEIGHTED_STOUFFERS = BibTeX(
    """
    @article{zaykin2011optimally,
      title={Optimally weighted Z-test is a powerful method for
             combining probabilities in meta-analysis},
      author={Zaykin, Dmitri V},
      journal={Journal of evolutionary biology},
      volume={24},
      number={8},
      pages={1836--1841},
      year={2011},
      publisher={Wiley Online Library}
      }
    """
)

CBP = Doi("10.1002/hbm.22138")

MAMP = Doi("10.1016/j.neuroimage.2015.08.027")

MAPBOT = Doi("10.1016/j.neuroimage.2017.06.032")

T2Z_TRANSFORM = BibTeX(
    """
    @article{hughett2007accurate,
      title={Accurate Computation of the F-to-z and t-to-z Transforms
             for Large Arguments},
      author={Hughett, Paul and others},
      journal={Journal of Statistical Software},
      volume={23},
      number={1},
      pages={1--5},
      year={2007},
      publisher={Foundation for Open Access Statistics}
    }
    """
)

T2Z_IMPLEMENTATION = Doi("10.5281/zenodo.32508")

LANCASTER_TRANSFORM = Doi("10.1002/hbm.20345")

LANCASTER_TRANSFORM_VALIDATION = Doi("10.1016/j.neuroimage.2010.02.048")

META_CLUSTER = Doi("10.1016/j.neuroimage.2015.06.044")

META_CLUSTER2 = Doi("10.1162/netn_a_00050")

META_ICA = Doi("10.1162/jocn_a_00077")

META_ICA2 = Doi("10.1162/jocn_a_00077")



"""Test nimare.extract."""
import os
from glob import glob

import nimare


def test_fetch_neurosynth(tmp_path_factory):
    """Smoke test for extract.fetch_neurosynth.

    Taken from the Neurosynth Python package.
    """
    tmpdir = tmp_path_factory.mktemp("test_fetch_neurosynth")
    url = (
        "https://raw.githubusercontent.com/neurosynth/neurosynth/master/neurosynth/tests/"
        "data/test_data.tar.gz"
    )
    nimare.extract.fetch_neurosynth(tmpdir, url=url, unpack=True)
    files = glob(os.path.join(tmpdir, "*.txt"))
    assert len(files) == 2



"""Utility functions for ontology tools."""
import numpy as np
import pandas as pd


def _generate_weights(rel_df, weights):
    """Create an IDxID DataFrame linking weight value to each relationship type.

    Parameters
    ----------
    rel_df : (X x 3) :obj:`pandas.DataFrame`
        DataFrame with three columns: input, output, and rel_type
        (relationship type).
    weights : :obj:`dict`
        Dictionary defining relationship weights. Each relationship type is a
        key and the associated value is the weight to use for that kind of
        relationship.

    Returns
    -------
    expanded_df : :obj:`pandas.DataFrame`
        Square DataFrame where rows correspond to input items, columns
        correspond to output items, and cells have the weights associated with
        the particular input/output relationship.

    Notes
    -----
    For example, if weights is {'partOf': 1}, the resulting expanded_df will
    have a value of 1 for all cells where the input item (row) is a part of the
    output item (column), and will have zeroes for all other cells.
    """
    # Override isSelf weight
    weights["isSelf"] = 1

    # Hierarchical expansion
    def get_weight(rel_type):
        weight = weights.get(rel_type, 0)
        return weight

    t_df = rel_df.copy()
    t_df["rel_type"] = t_df["rel_type"].apply(get_weight)
    weights_df = t_df.pivot_table(
        index="input", columns="output", values="rel_type", aggfunc=np.max
    )
    weights_df = weights_df.fillna(0)
    out_not_in = list(set(t_df["output"].values) - set(t_df["input"].values))
    in_not_out = list(set(t_df["input"].values) - set(t_df["output"].values))

    new_cols = pd.DataFrame(
        columns=in_not_out,
        index=weights_df.index,
        data=np.zeros((weights_df.shape[0], len(in_not_out))),
    )
    weights_df = pd.concat((weights_df, new_cols), axis=1)
    new_rows = pd.DataFrame(
        columns=weights_df.columns,
        index=out_not_in,
        data=np.zeros((len(out_not_in), weights_df.shape[1])),
    )
    weights_df = pd.concat((weights_df, new_rows), axis=0)
    all_cols = sorted(weights_df.columns.tolist())
    weights_df = weights_df.loc[all_cols, :]
    weights_df = weights_df.loc[:, all_cols]

    # expanding the hierarchical expansion to all related terms
    # this way, a single dot product will apply counts to all layers
    expanded_df = weights_df.copy()
    mat = weights_df.values

    for i, val in enumerate(weights_df.index):
        row = np.zeros((1, weights_df.shape[0]))
        row[0, i] = 1  # identity
        temp = np.zeros((1, weights_df.shape[0]))

        while not np.array_equal(temp != 0, row != 0):
            temp = np.copy(row)
            row = np.dot(row, mat)

            # Constrain weights to <=1.
            # Hopefully this won't mess with weights <1,
            # but will also prevent weights from adding to one another.
            row[row > 1] = 1
        expanded_df.loc[val] = np.squeeze(row)
    return expanded_df



"""Dataset and trained model downloading functions."""
import warnings

from . import utils
from .extract import (
    download_abstracts,
    download_cognitive_atlas,
    download_mallet,
    download_nidm_pain,
    download_peaks2maps_model,
    fetch_neurosynth,
)

__all__ = [
    "download_nidm_pain",
    "download_mallet",
    "download_cognitive_atlas",
    "download_abstracts",
    "download_peaks2maps_model",
    "fetch_neurosynth",
    "utils",
]

warnings.simplefilter("default")

warnings.warn(
    "{} is an experimental module under active development; use it at your "
    "own risk.".format(__name__),
    ImportWarning,
)



"""Test nimare.meta.ibma (image-based meta-analytic estimators)."""
import logging
import os.path as op
from contextlib import ExitStack as does_not_raise

import pytest
from nilearn.input_data import NiftiLabelsMasker

import nimare
from nimare.correct import FDRCorrector, FWECorrector
from nimare.meta import ibma

from .utils import get_test_data_path


@pytest.mark.parametrize(
    "meta,meta_kwargs,corrector,corrector_kwargs",
    [
        pytest.param(
            ibma.Fishers,
            {},
            FDRCorrector,
            {"method": "indep", "alpha": 0.001},
            id="Fishers",
        ),
        pytest.param(
            ibma.Stouffers,
            {"use_sample_size": False},
            None,
            {},
            id="Stouffers",
        ),
        pytest.param(
            ibma.Stouffers,
            {"use_sample_size": True},
            None,
            {},
            id="Stouffers_weighted",
        ),
        pytest.param(
            ibma.WeightedLeastSquares,
            {"tau2": 0},
            None,
            {},
            id="WeightedLeastSquares",
        ),
        pytest.param(
            ibma.DerSimonianLaird,
            {},
            None,
            {},
            id="DerSimonianLaird",
        ),
        pytest.param(
            ibma.Hedges,
            {},
            None,
            {},
            id="Hedges",
        ),
        pytest.param(
            ibma.SampleSizeBasedLikelihood,
            {"method": "ml"},
            None,
            {},
            id="SampleSizeBasedLikelihood_ml",
        ),
        pytest.param(
            ibma.SampleSizeBasedLikelihood,
            {"method": "reml"},
            None,
            {},
            id="SampleSizeBasedLikelihood_reml",
        ),
        pytest.param(
            ibma.VarianceBasedLikelihood,
            {"method": "ml"},
            None,
            {},
            id="VarianceBasedLikelihood_ml",
        ),
        pytest.param(
            ibma.VarianceBasedLikelihood,
            {"method": "reml"},
            None,
            {},
            id="VarianceBasedLikelihood_reml",
        ),
        pytest.param(
            ibma.PermutedOLS,
            {"two_sided": True},
            FWECorrector,
            {"method": "montecarlo", "n_iters": 100, "n_cores": 1},
            id="PermutedOLS",
        ),
    ],
)
def test_ibma_smoke(testdata_ibma, meta, meta_kwargs, corrector, corrector_kwargs):
    """Smoke test for IBMA estimators."""
    meta = meta(**meta_kwargs)
    res = meta.fit(testdata_ibma)
    z_img = res.get_map("z")
    assert isinstance(meta.results, nimare.results.MetaResult)
    assert isinstance(res, nimare.results.MetaResult)
    assert res.get_map("z", return_type="array").ndim == 1
    assert z_img.ndim == 3
    assert z_img.shape == (10, 10, 10)
    if corrector:
        corr = corrector(**corrector_kwargs)
        cres = corr.transform(res)
        assert cres.get_map("z", return_type="array").ndim == 1
        assert cres.get_map("z").ndim == 3


@pytest.mark.parametrize(
    "estimator,expectation,masker_source",
    [
        (ibma.Fishers, "error", "estimator"),
        (ibma.Stouffers, "error", "estimator"),
        (ibma.WeightedLeastSquares, "warning", "estimator"),
        (ibma.DerSimonianLaird, "warning", "estimator"),
        (ibma.Hedges, "warning", "estimator"),
        (ibma.SampleSizeBasedLikelihood, "no warning", "estimator"),
        (ibma.VarianceBasedLikelihood, "warning", "estimator"),
        (ibma.PermutedOLS, "no warning", "estimator"),
    ],
)
def test_ibma_with_custom_masker(testdata_ibma, caplog, estimator, expectation, masker_source):
    """Ensure voxel-to-ROI reduction works, but only for Estimators that allow it.

    Notes
    -----
    Currently masker_source is not used, but ultimately we will want to test cases where the
    Dataset uses a NiftiLabelsMasker.
    """
    atlas = op.join(get_test_data_path(), "test_pain_dataset", "atlas.nii.gz")
    masker = NiftiLabelsMasker(atlas)

    dset = testdata_ibma
    meta = estimator(mask=masker)

    if expectation == "error":
        with pytest.raises(ValueError):
            meta.fit(dset)
    elif expectation == "warning":
        with caplog.at_level(logging.WARNING, logger="nimare.meta.ibma"):
            meta.fit(dset)
            assert "will likely produce biased results" in caplog.text
        caplog.clear()
    else:
        with caplog.at_level(logging.WARNING, logger="nimare.meta.ibma"):
            meta.fit(dset)
            assert "will likely produce biased results" not in caplog.text
        caplog.clear()

    # Only fit the estimator if it doesn't raise a ValueError
    if expectation != "error":
        assert isinstance(meta.results, nimare.results.MetaResult)
        assert meta.results.maps["z"].shape == (5,)
        assert meta.results.get_map("z").shape == (10, 10, 10)


@pytest.mark.parametrize(
    "resample,resample_kwargs,expectation",
    [
        (False, {}, pytest.raises(ValueError)),
        (None, {}, pytest.raises(ValueError)),
        (True, {}, does_not_raise()),
        (
            True,
            {"resample__clip": False, "resample__interpolation": "continuous"},
            does_not_raise(),
        ),
    ],
)
def test_ibma_resampling(testdata_ibma_resample, resample, resample_kwargs, expectation):
    """Test image-based resampling performance."""
    meta = ibma.Fishers(resample=resample, **resample_kwargs)
    with expectation:
        meta.fit(testdata_ibma_resample)
    if isinstance(expectation, does_not_raise):
        assert isinstance(meta.results, nimare.results.MetaResult)



"""Workflow for running a SCALE meta-analysis from a Sleuth text file."""
import logging
import os
import pathlib
from shutil import copyfile

import numpy as np

from ..dataset import Dataset
from ..io import convert_sleuth_to_dataset
from ..meta import SCALE

LGR = logging.getLogger(__name__)


def scale_workflow(
    dataset_file,
    baseline=None,
    output_dir=None,
    prefix=None,
    n_iters=2500,
    v_thr=0.001,
    n_cores=-1,
):
    """Perform SCALE meta-analysis from Sleuth text file or NiMARE json file.

    Warnings
    --------
    This method is not yet implemented.
    """
    if dataset_file.endswith(".json"):
        dset = Dataset(dataset_file, target="mni152_2mm")
    elif dataset_file.endswith(".txt"):
        dset = convert_sleuth_to_dataset(dataset_file, target="mni152_2mm")
    else:
        dset = Dataset.load(dataset_file)

    boilerplate = """
A specific coactivation likelihood estimation (SCALE; Langner et al., 2014)
meta-analysis was performed using NiMARE. The input dataset included {n}
studies/experiments.

Voxel-specific null distributions were generated using base rates from {bl}
with {n_iters} iterations. Results were thresholded at p < {thr}.

References
----------
- Langner, R., Rottschy, C., Laird, A. R., Fox, P. T., & Eickhoff, S. B. (2014).
Meta-analytic connectivity modeling revisited: controlling for activation base
rates. NeuroImage, 99, 559-570.
    """
    boilerplate = boilerplate.format(
        n=len(dset.ids),
        thr=v_thr,
        bl=baseline if baseline else "a gray matter template",
        n_iters=n_iters,
    )

    # At the moment, the baseline file should be an n_coords X 3 list of matrix
    # indices matching the dataset template, where the base rate for a given
    # voxel is reflected by the number of times that voxel appears in the array
    if not baseline:
        ijk = np.vstack(np.where(dset.masker.mask_img.get_fdata())).T
    else:
        ijk = np.loadtxt(baseline)

    estimator = SCALE(ijk=ijk, n_iters=n_iters, n_cores=n_cores)
    estimator.fit(dset)

    if output_dir is None:
        output_dir = os.path.dirname(dataset_file)
    else:
        pathlib.Path(output_dir).mkdir(parents=True, exist_ok=True)

    if prefix is None:
        base = os.path.basename(dataset_file)
        prefix, _ = os.path.splitext(base)
        prefix += "_"
    elif not prefix.endswith("_"):
        prefix = prefix + "_"

    estimator.results.save_maps(output_dir=output_dir, prefix=prefix)
    copyfile(dataset_file, os.path.join(output_dir, prefix + "input_coordinates.txt"))

    LGR.info("Workflow completed.")
    LGR.info(boilerplate)



"""Test nimare.annotate.gclda (GCLDA)."""
import nibabel as nib
import numpy as np
import pandas as pd
import pytest

from nimare import annotate, decode


def test_gclda_symmetric(testdata_laird):
    """A smoke test for GCLDA with symmetric regions."""
    counts_df = annotate.text.generate_counts(
        testdata_laird.texts,
        text_column="abstract",
        tfidf=False,
        min_df=1,
        max_df=1.0,
    )

    with pytest.raises(ValueError):
        annotate.gclda.GCLDAModel(
            counts_df,
            testdata_laird.coordinates,
            mask=testdata_laird.masker.mask_img,
            n_regions=3,
            symmetric=True,
        )

    model = annotate.gclda.GCLDAModel(
        counts_df,
        testdata_laird.coordinates,
        mask=testdata_laird.masker.mask_img,
        n_regions=2,
        symmetric=True,
    )
    model.fit(n_iters=5, loglikely_freq=5)

    # Create ROI to decode
    arr = np.zeros(testdata_laird.masker.mask_img.shape, int)
    arr[40:44, 45:49, 40:44] = 1
    mask_img = nib.Nifti1Image(arr, testdata_laird.masker.mask_img.affine)
    decoded_df, _ = decode.discrete.gclda_decode_roi(model, mask_img)
    assert isinstance(decoded_df, pd.DataFrame)

    # Decode the ROI as a continuous map
    decoded_df, _ = decode.continuous.gclda_decode_map(model, mask_img)
    assert isinstance(decoded_df, pd.DataFrame)

    # Encode text
    encoded_img, _ = decode.encode.gclda_encode(model, "fmri activation")
    assert isinstance(encoded_img, nib.Nifti1Image)


def test_gclda_asymmetric(testdata_laird):
    """A smoke test for GCLDA with three asymmetric regions."""
    counts_df = annotate.text.generate_counts(
        testdata_laird.texts,
        text_column="abstract",
        tfidf=False,
        min_df=1,
        max_df=1.0,
    )
    model = annotate.gclda.GCLDAModel(
        counts_df,
        testdata_laird.coordinates,
        mask=testdata_laird.masker.mask_img,
        n_regions=3,
        symmetric=False,
    )
    model.fit(n_iters=5, loglikely_freq=5)

    # Create ROI to decode
    arr = np.zeros(testdata_laird.masker.mask_img.shape, int)
    arr[40:44, 45:49, 40:44] = 1
    mask_img = nib.Nifti1Image(arr, testdata_laird.masker.mask_img.affine)
    decoded_df, _ = decode.discrete.gclda_decode_roi(model, mask_img)
    assert isinstance(decoded_df, pd.DataFrame)

    # Decode the ROI as a continuous map
    decoded_df, _ = decode.continuous.gclda_decode_map(model, mask_img)
    assert isinstance(decoded_df, pd.DataFrame)

    # Encode text
    encoded_img, _ = decode.encode.gclda_encode(model, "fmri activation")
    assert isinstance(encoded_img, nib.Nifti1Image)
